from leapp.actors import Actor
from leapp.exceptions import StopActorExecutionError
from leapp.tags import FirstPhaseTag, WorkflowApiTestWorkflowTag
from leapp.workflows.api.v3 import DepCheckAPI1, DepCheckAPI2, DepCheckAPI3, DepCheckAPI4


class DepCheck(Actor):
    """
    This actor verifies the Depedencies APIs to be working as intended.

    DepCheckAPI1 and DepCheckAPI3 consume DepCheck1 and DepCheck3 messages.
    DepCheckAPI2 and DepCheckAPI4 produce DepCheck2 and DepCheck4 messages.

    DepCheckAPI4 requires DepCheckAPI3
    DepCheckAPI3 requires DepCheckAPI2
    DepCheckAPI2 requires DepCheckAPI1

    Therefore by requiring DepCheckAPI4 automatically all 4 DepCheckAPI variants
    are used to amend the consumes and produces fields of the actor.

    Please Note: This is purely a test to ensure that this is the case.
    I highly recommend instead of relying on the dependencies, that each actor
    exactly writes what APIs it uses.
    """

    name = 'dep_check'
    consumes = ()
    produces = ()
    apis = (DepCheckAPI4,)  # A no no normally, be explicit what you use, however for this test it's necessary
    tags = (FirstPhaseTag, WorkflowApiTestWorkflowTag)

    def process(self):
        api1 = DepCheckAPI1()
        if len(api1.consume()) != 1:
            raise StopActorExecutionError(
                "Expected 1 result from DepCheckAPI1.consume => {}".format(len(api1.consume())))

        api2 = DepCheckAPI2()
        api2.produce()

        api3 = DepCheckAPI3()
        if len(api3.consume()) != 2:
            raise StopActorExecutionError(
                "Expected 2 results from DepCheckAPI3.consume => {}".format(len(api3.consume())))

        api4 = DepCheckAPI4()
        api4.produce()



from leapp.models import Model, fields
from leapp.topics import ApiTestTopic


class ApiTest(Model):
    topic = ApiTestTopic

    data = fields.String()


class ApiTestProduce(ApiTest):
    pass


class ApiTestConsume(ApiTest):
    pass



import contextlib
import linecache
import logging
import os
import pkgutil
import sys
import warnings
from io import UnsupportedOperation
from multiprocessing import Process, Queue

import leapp.libraries.actor  # noqa # pylint: disable=unused-import
from leapp.actors import get_actor_metadata, get_actors
from leapp.exceptions import (ActorInspectionFailedError, LeappRuntimeError, MultipleActorsError,
                              UnsupportedDefinitionKindError)
from leapp.repository.definition import DefinitionKind
from leapp.utils.audit import create_audit_entry
from leapp.utils.deprecation import _LeappDeprecationWarning
from leapp.utils.libraryfinder import LeappLibrariesFinder


def inspect_actor(definition, result_queue):
    """
    Retrieves the actor information in a child process and returns the results back through `result_queue`.

    :param definition: the actor definition to load
    :type definition: :py:class:`ActorDefinition`
    :param result_queue: queue to pass results back to the calling process
    :type result_queue: :py:class:`multiprocessing.Queue`
    """
    definition.load()
    result = [get_actor_metadata(actor) for actor in get_actors()]
    result = [entry for entry in result if entry['path'] in definition.full_path]
    result_queue.put(result)


class ActorCallContext(object):
    """
    Wraps the actor execution into child process.
    """

    def __init__(self, definition, logger, messaging, config_model, skip_dialogs):
        """
        :param definition: Actor definition
        :type definition: :py:class:`leapp.repository.actor_definition.ActorDefinition`
        :param logger: Logger
        :type logger: :py:class:`logging.Logger`
        :param messaging: Leapp Messaging
        :type messaging: :py:class:`leapp.messaging.BaseMessaging`
        :param config_model: Workflow provided configuration model
        :type config_model: :py:class:`leapp.models.Model` derived class
        """
        self.definition = definition
        self.logger = logger
        self.messaging = messaging
        self.config_model = config_model
        self.skip_dialogs = skip_dialogs

    @staticmethod
    def _do_run(stdin, logger, messaging, definition, config_model, skip_dialogs, args, kwargs):
        if stdin is not None:
            try:
                sys.stdin = os.fdopen(stdin)
            except OSError:
                pass
        with warnings.catch_warnings(record=True) as recording:
            warnings.simplefilter(action="always", category=_LeappDeprecationWarning)
            definition.load()
            with definition.injected_context():
                target_actor = [actor for actor in get_actors() if actor.name == definition.name][0]
                actor_instance = target_actor(logger=logger, messaging=messaging, config_model=config_model,
                                              skip_dialogs=skip_dialogs)
                actor_instance.run(*args, **kwargs)
            try:
                # By this time this is no longer set, so we have to get it back
                os.environ['LEAPP_CURRENT_ACTOR'] = actor_instance.name
                for rec in recording:
                    if issubclass(rec.category, _LeappDeprecationWarning):
                        entry = {
                            'message': str(rec.message),
                            'filename': rec.filename,
                            'line': linecache.getline(rec.filename, rec.lineno) if rec.line is None else rec.line,
                            'lineno': rec.lineno,
                            'since': rec.category.since,
                            'reason': rec.category.msg
                        }
                        create_audit_entry('deprecation', entry)
            finally:
                # Remove it again
                os.environ.pop('LEAPP_CURRENT_ACTOR')

    def run(self, *args, **kwargs):
        """
        Performs the actor execution in the child process.
        """
        try:
            stdin = sys.stdin.fileno()
        except UnsupportedOperation:
            stdin = None
        p = Process(target=self._do_run,
                    args=(stdin, self.logger, self.messaging, self.definition, self.config_model,
                          self.skip_dialogs, args, kwargs))
        p.start()
        p.join()
        if p.exitcode != 0:
            raise LeappRuntimeError(
                'Actor {actorname} unexpectedly terminated with exit code: {exitcode}'
                .format(actorname=self.definition.name, exitcode=p.exitcode))


class ActorDefinition(object):
    """
    Defines actor resources.

    """

    def __init__(self, directory, repo_dir, log=None):
        """
        :param log: Logger
        :type log: :py:class:`logging.Logger`
        :param directory: Actor directory
        :type directory: str
        :param repo_dir: Repository directory
        :type repo_dir: str
        """
        self.log = log or logging.getLogger('leapp.actor')
        self._directory = directory
        self._repo_dir = repo_dir
        self._definitions = {}
        self._module = None
        self._discovery = None

    @property
    def full_path(self):
        return os.path.realpath(os.path.join(self._repo_dir, self._directory))

    def add(self, kind, path):
        """
        Adds any kind of actor resource to the Definition

        :param kind: kind of resource added
        :type kind: str
        :param path: path to the added resource
        :type path: str
        """
        if kind not in DefinitionKind.ACTOR_WHITELIST:
            self.log.error("Attempt to add item type %s to actor that is not supported", kind.name)
            raise UnsupportedDefinitionKindError('Actors do not support {kind}.'.format(kind=kind.name))
        self._definitions.setdefault(kind, []).append(path)

    def serialize(self):
        """
        :return: dump of actor resources (path, name, tools, files, libraries, tests)
        """
        return {
            'path': self.directory,
            'name': self.name,
            'class_name': self.class_name,
            'description': self.description,
            'tags': self.tags,
            'consumes': self.consumes,
            'produces': self.produces,
            'apis': self.apis,
            'dialogs': [dialog.serialize() for dialog in self.dialogs],
            'tools': self.tools,
            'files': self.files,
            'libraries': self.libraries,
            'tests': self.tests
        }

    def load(self):
        """
        Loads the actor module to be introspectable.
        """
        if not self._module:
            with self.injected_context():
                path = os.path.abspath(os.path.join(self._repo_dir, self.directory))
                for importer, name, is_pkg in pkgutil.iter_modules((path,)):
                    if not is_pkg:
                        self._module = importer.find_module(name).load_module(name)
                        break

    def discover(self):
        """
        Performs introspection through a subprocess.

        :return: Dictionary with discovered items.
        """
        if not self._discovery:
            self.log.debug("Starting actor discovery in %s", self.directory)
            q = Queue(1)
            p = Process(target=inspect_actor, args=(self, q))
            p.start()
            p.join()
            if p.exitcode != 0:
                self.log.error("Process inspecting actor in %s failed with %d", self.directory, p.exitcode)
                raise ActorInspectionFailedError('Inspection of actor in {path} failed'.format(path=self.directory))
            result = q.get()
            if not result:
                self.log.error("Process inspecting actor in %s returned no result", self.directory)
                raise ActorInspectionFailedError(
                    'Inspection of actor in {path} produced no results'.format(path=self.directory))
            if len(result) > 1:
                self.log.error("Actor in %s returned multiple actors", self.directory)
                raise MultipleActorsError(self.directory)
            self._discovery = result[0]
            for tag in self._discovery['tags']:
                if self not in tag.actors:
                    tag.actors += (self,)
        return self._discovery

    def __call__(self, messaging=None, logger=None, config_model=None, skip_dialogs=False):
        return ActorCallContext(definition=self, messaging=messaging, logger=logger, config_model=config_model,
                                skip_dialogs=skip_dialogs)

    @property
    def dialogs(self):
        """
        :return: Tuple of defined dialogs
        """
        return self.discover()['dialogs']

    @property
    def consumes(self):
        """
        :return: Tuple of consumed models
        """
        return self.discover()['consumes']

    @property
    def produces(self):
        """
        :return: Tuple of produced models
        """
        return self.discover()['produces']

    @property
    def tags(self):
        """
        :return: Tuple of tags assigned to the actor
        """
        return self.discover()['tags']

    @property
    def class_name(self):
        """
        :return: Actor class name
        """
        return self.discover()['class_name']

    @property
    def name(self):
        """
        :return: Actor internal name
        """
        return self.discover()['name']

    @property
    def description(self):
        """
        :return: Actor description
        """
        return self.discover()['description']

    @contextlib.contextmanager
    def injected_context(self):
        """
        Prepares the actor environment for running the actor.
        This includes injecting actor private libraries into :py:mod:`leapp.libraries.actor`
        and setting environment variables for private tools and files.

        :note: Use with caution.
        """
        # Backup of the path variable
        path_backup = os.environ.get('PATH', '')
        os.environ['PATH'] = ':'.join(path_backup.split(':') + list(
            os.path.join(self._repo_dir, self._directory, path) for path in self.tools))

        files_backup = os.environ.get('LEAPP_FILES', None)
        if self.files:
            os.environ['LEAPP_FILES'] = os.path.join(self._repo_dir, self._directory, self.files[0])

        tools_backup = os.environ.get('LEAPP_TOOLS', None)
        if self.tools:
            os.environ['LEAPP_TOOLS'] = os.path.join(self._repo_dir, self._directory, self.tools[0])

        sys.meta_path.append(
            LeappLibrariesFinder(
                module_prefix='leapp.libraries.actor',
                paths=[os.path.join(self._repo_dir, self.directory, x) for x in self.libraries]))

        previous_path = os.getcwd()
        os.chdir(os.path.join(self._repo_dir, self._directory))
        try:
            yield
        finally:
            os.chdir(previous_path)

            # Restoration of the PATH environment variable
            os.environ['PATH'] = path_backup
            # Restoration of the LEAPP_FILES environment variable
            if files_backup is not None:
                os.environ['LEAPP_FILES'] = files_backup
            else:
                os.environ.pop('LEAPP_FILES', None)

            if tools_backup is not None:
                os.environ['LEAPP_TOOLS'] = tools_backup
            else:
                os.environ.pop('LEAPP_TOOLS', None)

    @property
    def apis(self):
        """
        :return: names of APIs used by this actor
        """
        return tuple(self.discover()['apis'])

    @property
    def directory(self):
        """
        :return: The folder path of the actor
        """
        return self._directory

    @property
    def tools(self):
        """
        :return: Tuple with path to the tools folder of the actor, empty tuple if none
        """
        return tuple(self._definitions.get(DefinitionKind.TOOLS, ()))

    @property
    def libraries(self):
        """
        :return: Tuple with path to the libraries folder of the actor, empty tuple if none
        """
        return tuple(self._definitions.get(DefinitionKind.LIBRARIES, ()))

    @property
    def files(self):
        """
        :return: Tuple with path to the files folder of the actor, empty tuple if none
        """
        return tuple(self._definitions.get(DefinitionKind.FILES, ()))

    @property
    def tests(self):
        """
        :return: Tuple with path to the tests folder of the actor, empty tuple if none
        """
        return tuple(self._definitions.get(DefinitionKind.TESTS, ()))



import json
import os
import tempfile

import mock
import py
import pytest

from leapp.repository.scan import scan_repo


@pytest.fixture(scope='module')
def repository():
    repository_path = py.path.local(os.path.join(os.path.dirname(os.path.dirname(__file__)), 'data', 'workflow-tests'))
    with repository_path.as_cwd():
        repo = scan_repo('.')
        repo.load(resolve=True)
        yield repo


def test_workflow(repository):
    with tempfile.NamedTemporaryFile() as test_log_file:
        os.environ['LEAPP_TEST_EXECUTION_LOG'] = test_log_file.name
        workflow = repository.lookup_workflow('UnitTest')()
        assert not workflow.initial
        assert workflow.phase_actors
        assert not workflow.consumes
        assert not workflow.produces
        workflow.run(skip_dialogs=True)
        test_log_file.seek(0)
        order = [json.loads(line.decode('utf-8'))['class_name'] for line in test_log_file]
        assert order.pop(0) == 'FirstActor'
        assert tuple(sorted([order.pop(0), order.pop(0)])) == ('SecondActor', 'SecondCommonActor')
        assert tuple(sorted([order.pop(0), order.pop(0)])) == ('BeforeCommonThirdActor', 'BeforeThirdActor')
        assert order.pop(0) == 'ThirdActor'
        assert tuple(sorted([order.pop(0), order.pop(0)])) == ('AfterCommonThirdActor', 'AfterThirdActor')
        assert order.pop(0) == 'FourthActor'
        assert order.pop(0) == 'FifthActor'
        assert not order


def test_workflow_until_actor(repository):
    with tempfile.NamedTemporaryFile() as test_log_file:
        os.environ['LEAPP_TEST_EXECUTION_LOG'] = test_log_file.name
        workflow = repository.lookup_workflow('UnitTest')()
        workflow.run(context='unit-test-context', until_actor='ThirdActor', skip_dialogs=True)
        test_log_file.seek(0)
        order = [json.loads(line.decode('utf-8'))['class_name'] for line in test_log_file]
        assert order.pop(0) == 'FirstActor'
        assert tuple(sorted([order.pop(0), order.pop(0)])) == ('SecondActor', 'SecondCommonActor')
        assert tuple(sorted([order.pop(0), order.pop(0)])) == ('BeforeCommonThirdActor', 'BeforeThirdActor')
        assert order.pop(0) == 'ThirdActor'
        assert not order


def test_workflow_until_phase_main(repository):
    with tempfile.NamedTemporaryFile() as test_log_file:
        os.environ['LEAPP_TEST_EXECUTION_LOG'] = test_log_file.name
        workflow = repository.lookup_workflow('UnitTest')()
        workflow.run(context='unit-test-context', until_phase='ThirdPhase.main', skip_dialogs=True)
        test_log_file.seek(0)
        order = [json.loads(line.decode('utf-8'))['class_name'] for line in test_log_file]
        assert order.pop(0) == 'FirstActor'
        assert tuple(sorted([order.pop(0), order.pop(0)])) == ('SecondActor', 'SecondCommonActor')
        assert tuple(sorted([order.pop(0), order.pop(0)])) == ('BeforeCommonThirdActor', 'BeforeThirdActor')
        assert order.pop(0) == 'ThirdActor'
        assert not order


def test_workflow_until_phase_before(repository):
    with tempfile.NamedTemporaryFile() as test_log_file:
        os.environ['LEAPP_TEST_EXECUTION_LOG'] = test_log_file.name
        workflow = repository.lookup_workflow('UnitTest')()
        workflow.run(context='unit-test-context', until_phase='ThirdPhase.before', skip_dialogs=True)
        test_log_file.seek(0)
        order = [json.loads(line.decode('utf-8'))['class_name'] for line in test_log_file]
        assert order.pop(0) == 'FirstActor'
        assert tuple(sorted([order.pop(0), order.pop(0)])) == ('SecondActor', 'SecondCommonActor')
        assert tuple(sorted([order.pop(0), order.pop(0)])) == ('BeforeCommonThirdActor', 'BeforeThirdActor')
        assert not order


def test_workflow_until_phase_full(repository):
    with tempfile.NamedTemporaryFile() as test_log_file:
        os.environ['LEAPP_TEST_EXECUTION_LOG'] = test_log_file.name
        workflow = repository.lookup_workflow('UnitTest')()
        workflow.run(context='unit-test-context', until_phase='ThirdPhase', skip_dialogs=True)
        test_log_file.seek(0)
        order = [json.loads(line.decode('utf-8'))['class_name'] for line in test_log_file]
        assert order.pop(0) == 'FirstActor'
        assert tuple(sorted([order.pop(0), order.pop(0)])) == ('SecondActor', 'SecondCommonActor')
        assert tuple(sorted([order.pop(0), order.pop(0)])) == ('BeforeCommonThirdActor', 'BeforeThirdActor')
        assert order.pop(0) == 'ThirdActor'
        assert tuple(sorted([order.pop(0), order.pop(0)])) == ('AfterCommonThirdActor', 'AfterThirdActor')
        assert not order


def test_workflow_skip_phases_until(repository):
    with tempfile.NamedTemporaryFile() as test_log_file:
        os.environ['LEAPP_TEST_EXECUTION_LOG'] = test_log_file.name
        workflow = repository.lookup_workflow('UnitTest')()
        workflow.run(context='unit-test-context', skip_phases_until='ThirdPhase', skip_dialogs=True)
        test_log_file.seek(0)
        order = [json.loads(line.decode('utf-8'))['class_name'] for line in test_log_file]
        assert order.pop(0) == 'FourthActor'
        assert order.pop(0) == 'FifthActor'
        assert not order


def test_workflow_reboot(repository):
    with mock.patch('leapp.workflows.reboot_system') as reboot_system_function:
        with tempfile.NamedTemporaryFile() as test_log_file:
            os.environ['LEAPP_TEST_EXECUTION_LOG'] = test_log_file.name
            workflow = repository.lookup_workflow('UnitTest')()
            reboot_system_function.return_value = None
            fpidx = 0
            for idx, (phase, _before, _main, _after) in enumerate(workflow.phase_actors):
                if phase.name == 'first-phase':
                    fpidx = idx
                    break
            workflow.phase_actors[fpidx][0].flags.restart_after_phase = True
            workflow.run(skip_dialogs=True)
            workflow.phase_actors[fpidx][0].flags.restart_after_phase = False
            reboot_system_function.assert_called_once_with()
            test_log_file.seek(0)
            order = [json.loads(line.decode('utf-8'))['class_name'] for line in test_log_file]
            assert order.pop(0) == 'FirstActor'
            assert not order


def test_workflow_error_policy_fail_immediately(repository):
    with tempfile.NamedTemporaryFile() as test_log_file:
        os.environ['LEAPP_TEST_EXECUTION_LOG'] = test_log_file.name
        workflow = repository.lookup_workflow('UnitTest')()
        repository.lookup_actor('FirstActor').should_report_error = True
        os.environ['FirstActor-ReportError'] = '1'
        workflow.run(skip_dialogs=True)
        del os.environ['FirstActor-ReportError']

        test_log_file.seek(0)
        order = [json.loads(line.decode('utf-8'))['class_name'] for line in test_log_file]
        assert order.pop(0) == 'FirstActor'
        assert not order
        assert workflow.errors and len(workflow.errors) == 1


def test_workflow_error_policy_fail_phase(repository):
    with tempfile.NamedTemporaryFile() as test_log_file:
        os.environ['LEAPP_TEST_EXECUTION_LOG'] = test_log_file.name
        os.environ['BeforeThirdActor-ReportError'] = '1'
        os.environ['AfterThirdActor-ReportError'] = '1'
        workflow = repository.lookup_workflow('UnitTest')()
        repository.lookup_actor('FirstActor').should_report_error = True
        workflow.run(skip_dialogs=True)
        test_log_file.seek(0)
        order = [json.loads(line.decode('utf-8'))['class_name'] for line in test_log_file]
        assert order.pop(0) == 'FirstActor'
        assert tuple(sorted([order.pop(0), order.pop(0)])) == ('SecondActor', 'SecondCommonActor')
        assert tuple(sorted([order.pop(0), order.pop(0)])) == ('BeforeCommonThirdActor', 'BeforeThirdActor')
        assert order.pop(0) == 'ThirdActor'
        assert tuple(sorted([order.pop(0), order.pop(0)])) == ('AfterCommonThirdActor', 'AfterThirdActor')
        assert not order
        assert workflow.errors and len(workflow.errors) == 2


def test_workflow_render_dialogs(repository):
    workflow = repository.lookup_workflow('UnitTest')()
    with tempfile.NamedTemporaryFile(mode='w') as stdin_dialog:
        stdin_dialog.write('Yes\n')
        stdin_dialog.seek(0)
        with mock.patch('sys.stdin.fileno', return_value=stdin_dialog.fileno()):
            workflow.run(skip_dialogs=False)



from leapp.workflows.api import WorkflowAPI
from leapp.workflows.api import v3


class TestAPI(WorkflowAPI):
    """ Implementation of the v2.TestAPI which depends on the v3.TestAPI implementation """
    apis = (v3.TestAPI,)

    def __init__(self):
        # Instance of the v3 API
        self._impl = v3.TestAPI()

    def order_change(self, c, b, a):
        """
        Returns the value of a, b, c as a tuple as implemented in the v3 API
        """
        print("v2.TestAPI.order_change => " + str((a, b, c)))
        # Adjusts the order of parameters based on the V3 api needs and returns the value
        return self._impl.order_change(b, c, a)

    def survivor(self, message):
        """
        Prefixes message and returns it.
        """
        print("v2.TestAPI.survivor")
        # Calls the v3 implementation and returns the value
        return self._impl.survivor(message)



from helpers import make_repository_dir_fixture


# NOTE(ivasilev) Assigning a fixture generated that way to some variable is a necessary prerequisite for it to be
# discovered by pylint
repodir_fixture = make_repository_dir_fixture(name='repository_dir', scope='session')



from leapp.actors import Actor
from leapp.tags import ThirdPhaseTag, UnitTestWorkflowTag


class ThirdActor(Actor):
    name = 'third_actor'
    description = 'No description has been provided for the third_actor actor.'
    consumes = ()
    produces = ()
    tags = (ThirdPhaseTag, UnitTestWorkflowTag)

    def process(self):
        from leapp.libraries.common.test_helper import log_execution
        log_execution(self)



from leapp.actors import Actor
from leapp.tags import FifthPhaseTag, UnitTestWorkflowTag


class FifthActor(Actor):
    name = 'fifth_actor'
    description = 'No description has been provided for the fifth_actor actor.'
    consumes = ()
    produces = ()
    tags = (FifthPhaseTag, UnitTestWorkflowTag)

    def process(self):
        from leapp.libraries.common.test_helper import log_execution
        log_execution(self)



import functools
import os
import sys

from argparse import ArgumentParser, _SubParsersAction, RawDescriptionHelpFormatter

from leapp.exceptions import CommandDefinitionError, UsageError, CommandError


class _LeappArgumentParser(ArgumentParser):
    def print_help(self):
        super(_LeappArgumentParser, self).print_help(file=sys.stderr)

    def error(self, message):
        self.print_help()
        sys.stderr.write('error: %s\n' % message)
        sys.exit(2)


class _LeappHelpFormatter(RawDescriptionHelpFormatter):
    """
    Capitalizes section headings in the help output
    """
    def start_section(self, heading):
        return super(_LeappHelpFormatter, self).start_section(heading.capitalize())


class _SubParserActionOverride(_SubParsersAction):
    """
    This class implements a workaround for an issue fixed in 2.7.9.
    See https://bugs.python.org/issue9351

    TL;DR: Before, 2.7.9 argparse._SubParserAction did not propagate sub parser default values
    to the global namespace when they were already defined .
    This implementation is a workaround to override those values additionally.

    The additional code will not be executed if python 2.7.9 or higher is found.
    """
    def __call__(self, parser, namespace, values, option_string=None):
        super(_SubParserActionOverride, self).__call__(parser, namespace, values, option_string)
        if sys.version_info >= (2, 7, 9):
            return
        parser_name = values[0]
        arg_strings = values[1:]
        parser = self._name_parser_map[parser_name]
        sub_namespace, arg_strings = parser.parse_known_args(arg_strings, None)
        for key, value in vars(sub_namespace).items():
            setattr(namespace, key, value)


class Command(object):
    """
    Command implements a convenient command-based argument parsing the framework.
    """
    def __init__(self, name, target=None, help='', description=None):  # noqa; pylint: disable=redefined-builtin
        """
        :param name: Name of the sub command
        :type name: str
        :param target: Function called when the command is invoked
        :type target: Callable
        :param help: Shows a help message
        :type help: str
        :param description: Extended description of the command (the default is `help`)
        :type description: str
        """
        self.name = name
        self.help = help
        self.description = description or help
        self._sub_commands = {}
        self._options = []
        self.target = target
        self.parent = None
        self.parser = None

    def execute(self, version):
        """
        Entry point to the command execution. It is used for the main entry function of an application.

        :param version: Version string to display for --version calls
        :type version: str
        :return: None
        """
        parser = _LeappArgumentParser(prog=os.path.basename(sys.argv[0]), formatter_class=_LeappHelpFormatter)
        parser.register('action', 'parsers', _SubParserActionOverride)
        parser.add_argument('--version', action='version', version=version)
        parser.set_defaults(func=None)
        if self._sub_commands:
            s = parser.add_subparsers(title='Main commands', metavar='')
        else:
            s = parser
        self.apply_parser(s, parser=parser)
        # NOTE(ivasilev) a clumsy hack to make leapp output help information for the
        # exact subcommand that has been called.
        # The original problem was with parser.parse_args() referring to parser for
        # topmost leapp command and not the subcommand during error processing.
        # Any ideas and proposals on a more elegant solution are welcome.
        args, leftover = parser.parse_known_args()
        if leftover:
            cmdname = args.prog.split()[-1]
            # NOTE(ivasilev) No vetting of cmdname is necessary here as unless a proper leapp subcommand was
            # invocated parser.parse_known_args() would have thrown an exception
            self._sub_commands[cmdname].parser.error(leftover)
        args.func(args)

    def get_inheritable_options(self):
        """
        :return: Returns all options that are marked as 'inherit'
        """
        return [option for option in self._options if option[2].get('inherit')]

    def called(self, args):
        """
        The actual call is dispatched through this method. It ensures that the parent is also called
        to allow generic handling of some flags (especially inherited flags).

        :param args: Arguments object that is a result of the `argparse` commandline parser
        :type args: :py:class:`argparse.Namespace`
        :return: None
        """
        try:
            if self.parent:
                self.parent.called(args)

            if self.target:
                self.target(args)
        except UsageError as e:
            self.parser.print_help()
            self.parser.exit(status=2, message='\nUsageError: {message}\n'.format(message=e.message))
        except CommandError as e:
            self.parser.exit(status=2, message='\nError: {message}\n'.format(message=e.message))

    def apply_parser(self, sparser, parent=None, parser=None):
        """
        :param sparser: ArgumentParser.add_subparsers
        :type sparser: _SubParserActionOverride
        :param parent: Instance of :py:class:`_Command`
        :type parent: _Command
        :param parser: ArgumentParser instance usually received from sparser.add_parser
        :type parser: argparse.ArgumentParser
        :return: None
        """

        if self.parent is not parent:
            return

        if parser:
            self.parser = parser
        else:
            self.parser = sparser.add_parser(self.name, help=self.help, formatter_class=_LeappHelpFormatter,
                                             description='Description:\n\n' + self.description)

        self.parser.set_defaults(prog=self.parser.prog, func=self.called)
        self.parser.register('action', 'parsers', _SubParserActionOverride)
        inheritable = []
        current = self.parent
        while current:
            inheritable += current.get_inheritable_options()
            current = current.parent
        for args, kwargs, internal in self._options + inheritable:
            self.parser.add_argument(*args, **kwargs)

        if self._sub_commands:
            if not parser:
                subs = self.parser.add_subparsers(prog=self.parser.prog, title='Available subcommands',
                                                  help=self.help,  # noqa; pylint: disable=redefined-builtin
                                                  metavar='')
            else:
                subs = sparser
            for name, cmd in self._sub_commands.items():
                cmd.apply_parser(subs, parent=self)

    def add_sub(self, cmd):
        """
        Adds a sub command to this command

        :param cmd: The sub command object
        :type cmd: :py:class:`leapp.utils.clicmd.Command`
        :return: self
        """
        cmd.parent = self
        self._sub_commands[cmd.name] = cmd
        return self

    def __call__(self, *args, **kwargs):
        kwargs['parent'] = self
        return command(*args, **kwargs)

    def _add_opt(self, *args, **kwargs):
        internal = kwargs.pop('internal', {})
        self._options.append((args, kwargs, internal))

    def add_option(self, name, short_name='', help='',  # noqa; pylint: disable=redefined-builtin
                   is_flag=False, inherit=False, value_type=str, wrapped=None, action=None, metavar=None):
        """
        Add an option

        :param name: Name of the option
        :type name: str
        :param short_name: Short name of the option (one letter)
        :type short_name: str
        :param help: Help string for the option
        :type help: str
        :param is_flag: Decides if it is a flag
        :type is_flag: bool
        :param inherit: Decides if this option should be inherited by sub commands
        :type inherit: bool
        :param value_type: Type of the value by default string
        :param wrapped: Function that is wrapped (aka the target)
        :type wrapped: Callable
        :param action: ArgumentParser actions to take (e.g. store)
        :type action: str
        :param metavar: Changes the display name of arguments in generated help messages.
                        It has no influence on the attribute name from the generated arguments namespace.
        :type metavar: str
        :return: self
        """
        name = name.lstrip('-')
        names = ['--' + name]
        kwargs = {}
        if short_name:
            short_name = short_name.lstrip('-')
            if len(short_name) != 1:
                raise CommandDefinitionError("Short name should be one letter only")
            names.insert(0, '-' + short_name)
        if not action:
            action = 'store'
            if is_flag:
                action = 'store_true'
            elif value_type:
                kwargs['type'] = value_type
        if metavar:
            kwargs['metavar'] = metavar
        self._add_opt(*names, help=help,  # noqa; pylint: disable=redefined-builtin
                      action=action, internal={'wrapped': wrapped, 'inherit': inherit}, **kwargs)
        return self

    def add_argument(self, name, value_type=None, help='', wrapped=None):  # noqa; pylint: disable=redefined-builtin
        """

        :param name:
        :param value_type:
        :param help:
        :param wrapped:
        :return:
        """
        self._add_opt(name.replace('-', '_'), help=help,  # noqa; pylint: disable=redefined-builtin
                      type=value_type or str, internal={'wrapped': wrapped})
        return self


def command(name, help='', description=None, parent=None):  # noqa; pylint: disable=redefined-builtin
    """
    Decorator to mark a function as a sub command

    :param name: Sub command name
    :type name: str
    :param help: Help string for the sub command
    :type help: str
    :param description: Extended description for the sub command defaults to help if it is not set
    :type description: str
    :param parent: Instance to the parent command if it is a sub-command
    :type parent: Command
    """
    def wrapper(f):
        if not hasattr(f, 'command'):
            f.command = Command(name, help=help,  # noqa; pylint: disable=redefined-builtin
                                target=f, description=description)
        else:
            f.command.name = name
            f.command.help = help
            f.command.description = description or help
            f.command.target = f
        if parent:
            parent.add_sub(f.command)
        return f
    return wrapper


def _ensure_command(wrapped):
    @functools.wraps(wrapped)
    def wrapper(f):
        if not hasattr(f, 'command'):
            f.command = Command('')
        return wrapped(f)
    return wrapper


def command_arg(name, value_type=None, help=''):  # noqa; pylint: disable=redefined-builtin
    """
    Decorator wrapping functions to add command line arguments to the sub command to be invoked

    :param name: Name of the argument
    :param value_type: Type of the argument
    :param help: Help string for the argument
    """
    @_ensure_command
    def wrapper(f):
        f.command.add_argument(name, value_type=value_type, help=help,  # noqa; pylint: disable=redefined-builtin
                               wrapped=f)
        return f
    return wrapper


def command_opt(name, **kwargs):
    """
    Decorator wrapping functions to add command line options to the sub command to be invoked

    :param name: Name of the option
    :param kwargs: parameters as specified in :py:func:`leapp.utils.clicmd.Command.add_option`
    """
    @_ensure_command
    def wrapper(f):
        f.command.add_option(name, wrapped=f, **kwargs)
        return f
    return wrapper


def command_aware_wraps(f):
    """
    Decorator passing the command attribute of the wrapped function to the wrapper

    This needs to be used by decorators that are trying to wrap clicmd decorated command functions.
    """
    additional = ()
    if hasattr(f, 'command'):
        additional = ('command',)
    return functools.wraps(f, assigned=functools.WRAPPER_ASSIGNMENTS + additional)



class Dialog(object):
    """
    The dialog class is used to declare the information passed to the user and retrieved from the user
    during an interaction.
    """

    components = ()
    """ Components to display in the given order in the dialog """

    title = None
    """ Title of the dialog """

    reason = None
    """ An explanation for what the data in this dialog is needed for. """

    scope = None
    """ Unique scope identifier for the data to be stored in the answer files. Scope + component key is used to
        address values.
    """

    def __init__(self, scope, reason, title=None, components=None):
        """

        :param scope: Unique scope identifier for the data to be stored in the answer files. Scope + component key
                      is used to address values.
        :type scope: str
        :param reason: An explanation for what the data in this dialog is needed for.
        :type reason: str
        :param title: Title of the dialog
        :type title: str
        :param components: Components to display in the given order in the dialog
        :type components: tuple(leapp.dialogs.components.Component)
        """
        self.components = components or self.components
        self.title = title
        self.scope = scope
        self.reason = reason
        self._store = None
        self._min_label_width = None

    def serialize(self):
        """
        :return: Dictionary with the serialized representation of a component
        """
        return {
            'components': [component.serialize() for component in self.components],
            'title': self.title,
            'reason': self.reason,
            'scope': self.scope
        }

    @property
    def answerfile_sections(self):
        return {"{}.{}".format(self.scope, c.key): c.choices for c in self.components}

    @property
    def min_label_width(self):
        """
        :return: Returns the highest number of characters all labels in the dialog have, to help calculating the minimum
                 width the labels should have.
        """
        if not self._min_label_width:
            self._min_label_width = max(len(comp.label) for comp in self.components if comp.label)
        return self._min_label_width

    def answer(self, component, value):
        """
        Implements storing of answers.

        :param component: Component for which the answer is set
        :param value: The answer value
        :return: None
        """
        component.value = value
        self._store.answer(self.scope, component.key, value)

    def component_by_key(self, key):
        """
        Finds the component with the given key

        :param key: Key of the component to return
        :type key: str
        :return: Component found or None
        """
        for component in self.components:
            if component.key == key:
                return component
        return None

    def get_answers(self, store):
        """
        Checks answerstore if an answer is recorded for the dialog.

        :param store: AnswerStore instance
        :return: Dictionary with answers once retrieved
        """
        store.translate(self)
        return dict(store.get(self.scope, {}))

    def request_answers(self, store, renderer):
        """
        Same as get_answers but with interactive retrieval of the answer
        in case no recorded answer found in answerstore.

        :param store: AnswerStore instance
        :param renderer: Target renderer instance
        :return: Dictionary with answers once retrieved
        """
        store.translate(self)
        if any([component.value is None for component in self.components]):
            self._store = store
            renderer.render(self)
            self._store = None
        return dict(store.get(self.scope, {}))



from leapp.models.fields import ModelMisuseError


def init_from_tuple(model_class, field_names, data):
    """
    Create an instance of the given model class and use the data to initialize the fields with it.

    :param model_class: Model class to initialize with the given data
    :param field_names: Names of the fields to be mapped from the tuple
    :param data: Data to initialize the Model from
    :type data: dict
    :return: Instance of this class
    """
    if len(field_names) != len(data):
        raise ModelMisuseError(
            'Number of field names do not match number of tuple elements while initializing ' + model_class.__name__)
    return model_class(**dict(zip(field_names, data)))



import os

import sys

from leapp.utils.repository import requires_repository, make_class_name, find_repository_basedir
from leapp.utils.clicmd import command_arg, command_opt, command
from leapp.exceptions import CommandError


_LONG_DESCRIPTION = '''
Creates a new Model with all necessary boilerplate code in the current
repository.

For more information please consider reading the documentation at:
https://red.ht/leapp-docs
'''


@command('new-model', help='Creates a new model', description=_LONG_DESCRIPTION)
@command_opt('topic', help='Assigns the given topic to the model', metavar='TopicClassName')
@command_arg('model-name')
@requires_repository
def cli(args):
    model_name = args.model_name
    basedir = find_repository_basedir('.')

    basedir = os.path.join(basedir, 'models')
    if not os.path.isdir(basedir):
        os.mkdir(basedir)

    model_path = os.path.join(basedir, model_name.lower() + '.py')
    if os.path.exists(model_path):
        raise CommandError("File already exists: {}".format(model_path))
    topic_usage = 'None #  TODO: import appropriate topic and set it here'
    topic_import = ''
    if args.topic:
        topic_usage = args.topic
        topic_import = 'from leapp.topics import {}\n'.format(args.topic)

    with open(model_path, 'w') as f:
        f.write('''from leapp.models import Model, fields
{topic_import}

class {model_name}(Model):
    topic = {topic_usage}
'''.format(model_name=make_class_name(model_name), topic_import=topic_import, topic_usage=topic_usage))

    sys.stdout.write("New model {} has been created in {}\n".format(make_class_name(model_name),
                                                                    os.path.realpath(model_path)))



import multiprocessing

import six
from six.moves import configparser

from leapp.utils.audit import create_audit_entry


class AnswerStore(object):
    """
    AnswerStore handles storing and loading answer files for user questions.
    """

    def __init__(self, manager=None):
        """
        Initialize the answer store.
        :param manager: Passes through a given instance of multiprocessing.Manager() to use, by default it creates
                        an own instance.
        """
        self._manager = manager or multiprocessing.Manager()
        self._storage = self._manager.dict()

    def answer(self, scope, key, value):
        dialog_scope = self._storage.get(scope, {})
        dialog_scope[key] = value
        # As _storage is a dict proxy, we have to explicitly update the object via __setitem__
        # Otherwise we won't get the update propagated to the parent process
        # So don't even bother updating this to some more 'pythonic' coding style
        self._storage[scope] = dialog_scope

    def update(self, answer_file, allow_missing=False):
        """
        Update answerfile with all answers from answerstore that have correspondent sections in the file.

        Returns a list of sections that were not found in original answerfile and thus were not updated.
        """
        # NOTE(ivasilev): py2 configparser doesn't have any means to save comments in ini file. Switch to cfgparse?
        conf = configparser.SafeConfigParser(allow_no_value=True)
        conf.read(answer_file)
        not_updated = []
        for section, answerdict in self._storage.items():
            for opt, val in answerdict.items():
                if section not in conf.sections() and allow_missing:
                    conf.add_section(section)
                try:
                    conf.set(section, opt, str(val))
                except configparser.NoSectionError:
                    not_updated.append("{sec}.{opt}={val}".format(sec=section, opt=opt, val=val))
        with open(answer_file, 'w') as afile:
            conf.write(afile)
        return not_updated

    def load(self, answer_file):
        """
        Loads an answer file from the given location and updates the loaded data with it.

        :param answer_file: Path to the answer file to load.
        :return: None
        """
        conf = configparser.SafeConfigParser(allow_no_value=True)
        conf.read(answer_file)
        for section in conf.sections():
            self._storage[section] = self._manager.dict(conf.items(section=section, raw=True))

    def load_and_translate_for_workflow(self, answer_file, workflow):
        """
        Loads an answer file from the given location and updates the loaded data with it and translates the data to
        the correct value types based on the dialogs in the given workflow.

        :param answer_file: Path to the answer file to load.
        :param workflow:
        :return: None
        """
        conf = configparser.SafeConfigParser(allow_no_value=True)
        conf.read(answer_file)
        for section in conf.sections():
            self._storage[section] = dict(conf.items(section=section, raw=True))
        self.translate_for_workflow(workflow)

    def get(self, scope, fallback=None):
        """
        Dict compatible interface to get a sub dictionary by dialog scope.

        :param scope: Scope of the data to retrieve.
        :param fallback: Fallback value to return if not found.
        :return:
        """
        answer = self._storage.get(scope, fallback)
        create_audit_entry('dialog-answer', {'scope': scope, 'fallback': fallback, 'answer': answer})
        return answer

    def translate_for_workflow(self, workflow):
        """
        Translates the data for all dialogs in the current workflow.

        :param workflow: Instance of a workflow to translate all dialogs from.
        :type workflow: :py:class:`leapp.workflows.Workflow`
        :return: None
        """
        for dialog in workflow.dialogs:
            self.translate(dialog)

    def translate(self, dialog):
        """
        Translates configuration values from their string form to the corresponding value format based on the given
        dialog.

        :param dialog: A dialog instance to translate the data for.
        :type dialog: :py:class:`leapp.dialogs.dialog.Dialog`
        :return: None
        """
        entry = self._storage.get(dialog.scope)
        if entry:
            for component in dialog.components:
                if component.key in entry and isinstance(entry[component.key], six.string_types):
                    if component.value_type is bool:
                        entry[component.key] = entry[component.key].lower() == 'true'
                    elif component.value_type is int:
                        entry[component.key] = int(entry[component.key])
                    elif component.value_type is tuple:
                        elements = entry[component.key].split(';')
                        entry[component.key] = tuple(set(elements).intersection(component.choices))
                    elif hasattr(component, 'choices'):
                        value = entry[component.key]
                        entry[component.key] = value if value in component.choices else None
                    component.value = entry.get(component.key)
            self._storage.update({dialog.scope: entry})

    def generate(self, dialogs, answer_file_path):
        """
        Generates an answer file for the given dialogs and stores it to `answer_file_path`.
        :param dialogs:
        :param answer_file_path:
        :return:
        """
        with open(answer_file_path, 'w') as f:
            for dialog in dialogs:
                if not dialog.components:
                    # Skip dialogs without questions
                    continue

                f.writelines([
                    '[{}]\n'.format(dialog.scope),
                    '# {:<20}{}\n'.format('Title:', dialog.title),
                    '# {:<20}{}\n'.format('Reason:', dialog.reason)
                ])

                for component in dialog.components:
                    answer = self._storage.get(dialog.scope, {}).get(component.key)
                    default = component.default
                    choices = ''
                    answer_entry = ''
                    if hasattr(component, 'choices'):
                        choices += '# Available choices: {}\n'.format('/'.join(component.choices))
                        if component.value_type is tuple:
                            default = ';'.join(component.default)
                            answer = ';'.join(answer) if answer else answer
                            answer_entry = '#\n# Values are separated by semi-colon ";"\n'
                    if answer is not None:
                        answer_entry += '{key} = {value}\n'.format(key=component.key, value=answer)
                    else:
                        answer_entry += '# Unanswered question. Uncomment the following line with your answer\n'
                        answer_entry += '# {key} = {value}\n'.format(key=component.key, value=default or '')

                    f.writelines([
                        '# {}\n'.format(' {}.{} '.format(dialog.scope, component.key).center(77, '=')),
                        '# {:<20}{}\n'.format('Label:', component.label),
                        '# {:<20}{}\n'.format('Description:', component.description),
                        '# {:<20}{}\n'.format('Type:', component.value_type.__name__),
                        '# {:<20}{}\n'.format('Default:', default),
                        choices,
                        answer_entry,
                        '\n'
                    ])



"""
:py:mod:`leapp.libraries.common` represents an import location for shared libraries that
are placed in the repository's libraries folder.

Example:
      If any of the repositories has a libraries folder with a module.py python module, import it
      from the actor like this::

      from leapp.libraries.common import module

"""

LEAPP_BUILTIN_COMMON_INITIALIZED = False
"""Internal variable to the framework to inform if the common libraries have been already initialized"""



from __future__ import print_function
from multiprocessing import util, Process, Manager

import os

import leapp  # noqa: F401; pylint: disable=unused-import


def test_mp_is_patched():

    def child_fun(_lst):
        pid = os.fork()
        if not pid:
            os.execvpe('whatever', [], os.environ)
        else:
            os.wait()

    m = Manager()
    lst = m.list()

    p = Process(target=child_fun, args=(lst,))
    p.start()
    p.join()
    if lst:
        for el in lst:
            print(el)


def test_mp_workaround_applied():
    if getattr(util, 'os', None) is None:
        assert util.Finalize.__name__ == 'FixedFinalize'



from argparse import Namespace
from multiprocessing import Process

import mock
import pytest

from helpers import make_repository_dir_fixture
from leapp.exceptions import LeappRuntimeError, RepositoryConfigurationError
from leapp.repository.scan import find_and_scan_repositories, scan_repo
from leapp.snactor.commands.new_actor import cli as new_actor_cmd
from leapp.snactor.commands.new_tag import cli as new_tag_cmd
from leapp.snactor.commands.workflow.new import cli as new_workflow_cmd
import leapp.tags


# override repository_dir fixture generally defined in session scope
repository_test_repository_dir = make_repository_dir_fixture('repository_dir', scope='module')
empty_repodir_fixture = make_repository_dir_fixture(name='empty_repository_dir', scope='module')


def test_empty_repo(empty_repository_dir):
    with empty_repository_dir.as_cwd():
        repo = scan_repo(empty_repository_dir.strpath)
        repo.load(resolve=True)
        assert not repo.actors
        assert not repo.files
        assert not repo.models
        assert not repo.tags
        assert not repo.topics
        assert not repo.workflows
        assert not repo.lookup_workflow('Any')
        assert not repo.lookup_actor('Any')


def setup_repo(repository_dir):
    with repository_dir.as_cwd():
        new_tag_cmd(Namespace(tag_name='Test'))
        new_workflow_cmd(Namespace(name='Test', class_name=None, short_name=None))
        actor_path = new_actor_cmd(Namespace(
            actor_name='TestActor',
            tag=['TestTag', 'TestWorkflowTag'],
            consumes=[],
            produces=[],
        ))
        type(repository_dir)(actor_path).join('tests', 'test_this_actor.py').write('print("I am a test")')
        type(repository_dir)(actor_path).mkdir('libraries').mkdir('lib').join('__init__.py').write(
            '''from subprocess import call
# This is to ensure that actor tools are available on actor library load time
assert call(['woot-tool']) == 42

# This is to ensure that common tools are available on actor library load time
assert call(['common-woot-tool']) == 42

def do():
    # This must always fail - This function is crashing the actor ;-)
    assert call(['woot-tool']) == 0
        ''')
        repository_dir.mkdir('libraries').mkdir('lib').join('__init__.py').write(
            '''from subprocess import call
# This is to ensure that common tools are available on common library load time
assert call(['common-woot-tool']) == 42
        ''')
        type(repository_dir)(actor_path).mkdir('files').join('test.data').write('data')
        repository_dir.mkdir('files').join('common-test.data').write('data')
        tool_path = type(repository_dir)(actor_path).mkdir('tools').join('woot-tool')
        woot_tool_content = '''#!/bin/bash
echo 'WOOT'
exit 42
'''
        tool_path.write(woot_tool_content)
        tool_path.chmod(0o755)
        tool_path = repository_dir.mkdir('tools').join('common-woot-tool')
        tool_path.write(woot_tool_content)
        tool_path.chmod(0o755)
        actor_file = type(repository_dir)(actor_path).join('actor.py')
        actor_content = actor_file.read().replace('pass', '''from leapp.libraries.actor.lib import do
        import leapp.libraries.common.lib

        do()''')
        actor_file.write(actor_content)


def test_repo(repository_dir):
    setup_repo(repository_dir)

    def _run_test(repo_path):
        with repo_path.as_cwd():
            repository = find_and_scan_repositories(repo_path.dirpath().strpath)
            assert repository
            repository.load(resolve=True)
            assert getattr(leapp.tags, 'TestTag')
            assert repository.lookup_actor('TestActor')
            assert repository.lookup_workflow('TestWorkflow')
            assert not repository.lookup_workflow('MissingWorkflow')
            assert not repository.lookup_actor('MissingActor')
            assert repository.repos
            assert len(repository.serialize()) >= 1
            assert repository.actors
            assert not repository.topics
            assert not repository.models
            assert repository.tags
            assert repository.workflows
            assert repository.tools
            assert repository.libraries
            assert repository.files
            with pytest.raises(LeappRuntimeError):
                repository.lookup_actor('TestActor')().run()

    p = Process(target=_run_test, args=(repository_dir,))
    p.start()
    p.join()
    assert p.exitcode == 0


def test_find_and_scan_repositories_no_user_repos_config(empty_repository_dir):
    repo_path = empty_repository_dir.dirpath().strpath
    with mock.patch('leapp.repository.manager.RepositoryManager.get_missing_repo_links', return_value={'42'}):
        with mock.patch('leapp.utils.repository.get_global_repositories_data', return_value={}):
            with mock.patch('leapp.utils.repository.get_user_config_repos', return_value='/no/such/file.json'):
                with pytest.raises(RepositoryConfigurationError) as err:
                    find_and_scan_repositories(repo_path, include_locals=True)
                assert 'No repos configured' in err.value.message



from leapp.actors import Actor
from leapp.libraries.common.deprecation import (DeprecatedNoInit, DeprecatedNoInitDerived, DeprecatedWithInit,
                                                DeprecatedWithInitDerived, deprecated_function)
from leapp.models import DeprecatedModel, SuppressedDeprecatedModel
from leapp.tags import DeprecationPhaseTag, DeprecationWorkflowTag
from leapp.utils.deprecation import suppress_deprecation, deprecated


def whatever(f):
    def wrapper():
        f()
    return wrapper


@deprecated(since='2011-11-11', message='never to be seen')
@whatever
def foobar():
    pass


@suppress_deprecation(SuppressedDeprecatedModel, foobar)
def test_fun(self):
    self.produce(SuppressedDeprecatedModel())
    foobar()


class DeprecationTests(Actor):
    """
    No documentation has been provided for the deprecation_tests actor.
    """

    name = 'deprecation_tests'
    consumes = ()
    produces = (DeprecatedModel, SuppressedDeprecatedModel)
    tags = (DeprecationWorkflowTag, DeprecationPhaseTag)

    def process(self):
        test_fun(self)
        deprecated_function()
        self.produce(DeprecatedModel())
        DeprecatedNoInit()
        DeprecatedNoInitDerived()
        DeprecatedWithInit()
        DeprecatedWithInitDerived()



from __future__ import print_function
import sys

from leapp.exceptions import LeappError, CommandError
from leapp.logger import configure_logger
from leapp.repository.scan import find_and_scan_repositories
from leapp.snactor.commands.workflow import workflow
from leapp.utils.clicmd import command_arg
from leapp.utils.output import Color
from leapp.utils.repository import requires_repository, find_repository_basedir

_DESCRIPTION = 'The following messages are attempted to be consumed before they are produced: {}'
_LONG_DESCRIPTION = '''
Perform workflow sanity checks

- check whether there is a message in the given workflow which is attempted to be consumed before it was produced

For more information please consider reading the documentation at:
https://red.ht/leapp-docs
'''


def print_fail(error):
    print('{red}FAIL: {error}{reset}'.format(red=Color.red, error=error, reset=Color.reset), file=sys.stderr, end='\n')


@workflow.command('sanity-check', help='Perform workflow sanity checks', description=_LONG_DESCRIPTION)
@command_arg('name')
@requires_repository
def cli(params):
    configure_logger()
    repository = find_and_scan_repositories(find_repository_basedir('.'), include_locals=True)
    try:
        repository.load()
    except LeappError as exc:
        sys.stderr.write(exc.message)
        sys.stderr.write('\n')
        sys.exit(1)

    wf = repository.lookup_workflow(params.name)
    if not wf:
        raise CommandError('Could not find any workflow named "{}"'.format(params.name))

    instance = wf()
    produced_late = set(instance.initial).intersection(set(instance.produces))
    if produced_late:
        print_fail(_DESCRIPTION.format(' '.join([m.__name__ for m in produced_late])))
        sys.exit(1)



import datetime
import functools
import inspect
import warnings
from collections import namedtuple

from leapp.models import Model


_suppressed_deprecations = set()


def suppress_deprecation(*suppressed_items):
    """
    Decorator to suppress deprecation warnings during the execution of the decorated entity.

    This can be used directly on Actor classes OR functions. This relies on the fact that a class has a `process`
    method, if the method does not exist, this may cause mayhem.

    :param suppressed_items: Variable number of arguments each for entities to suppress the deprecation warnings for.
    """
    def decorator(item):
        target_item = item
        if inspect.isclass(item):
            target_item = getattr(item, 'process', None)

        @functools.wraps(target_item)
        def process_wrapper(*args, **kwargs):
            for suppressed in suppressed_items:
                _suppressed_deprecations.add(suppressed)
            try:
                return target_item(*args, **kwargs)
            finally:
                for suppressed in suppressed_items:
                    _suppressed_deprecations.remove(suppressed)
        if inspect.isclass(item):
            item.process = process_wrapper
            return item
        return process_wrapper
    return decorator


class _LeappDeprecationWarning(DeprecationWarning):
    since = None
    message = None


def deprecated(since, message, stack_level_offset=0):
    """
    `deprecated` is a decorator for items to mark them as being deprecated

    :param message: Information about why it was deprecated and potentially a pointer for an alternative
    :param since: Date since when this item is marked deprecated
    :type since: str in the form YYYY-MM-DD (strftime format: %Y-%m-%d)
    :param stack_level_offset: Number of stack levels to report above the reported stack level. This is mainly useful,
                               when using the decorator on a base class and ensuring the derived classes are going to
                               report deprecation usage at the place the derived class was intantiated and not in the
                               constructor of the derived class.
    :type stack_level_offset: int
    """
    # Ensure %Y-%m-%d format otherwise it will raise a Value error
    since = datetime.datetime.strptime(since, "%Y-%m-%d").date().strftime("%Y-%m-%d")

    def decorator(item):
        kind = 'function'
        if inspect.isclass(item):
            if issubclass(item, Model):
                kind = 'Model'
            else:
                kind = 'class'

        _since = since

        class _DeprecationWarningContext(_LeappDeprecationWarning):
            since = _since
            msg = message

        item.__deprecation__ = namedtuple('Deprecation', ('since', 'message'))(since=since, message=message)

        # Storing the result so we can suppress later the actual thing.
        # In case it is a class, the class itself is returned, so no problem here, however if we want to suppress
        # the usage of a function, we need to check against the wrapper
        result = None

        def do_warn():
            if result in _suppressed_deprecations:
                return

            # Inserts temporarily a filter rule
            warnings.simplefilter(action='always', category=_LeappDeprecationWarning)
            # Issue warning
            warnings.warn(message='Usage of deprecated {kind} "{name}"'.format(
                kind=kind, name=item.__name__), category=_DeprecationWarningContext, stacklevel=3 + stack_level_offset)
            # Pops temporarily inserted filtration rule
            warnings.filters.pop(0)

        if inspect.isclass(item):
            old_init = item.__init__

            @functools.wraps(item.__init__, assigned=('__name__', '__doc__'))
            def wrapper(*args, **kwargs):
                do_warn()
                return old_init(*args, **kwargs)
            item.__init__ = wrapper
            result = item
        else:
            @functools.wraps(item)
            def wrapper(*args, **kwargs):
                do_warn()
                return item(*args, **kwargs)
            result = wrapper
        return result
    return decorator



import pytest

from leapp.utils.meta import get_flattened_subclasses, with_metaclass


def test_with_metaclass():
    class MetaClassCreated(Exception):
        pass

    class TestMetaClass(type):
        def __new__(mcs, name, bases, attrs):  # noqa; pylint: disable=unused-argument
            raise MetaClassCreated()

    with pytest.raises(MetaClassCreated):
        class MetaClassUser(with_metaclass(TestMetaClass)):
            pass


def test_get_flattened_subclasses():
    class BaseClass(object):
        pass

    class SubClass(BaseClass):
        pass

    class SubSubClass(SubClass):
        pass

    class SubSubSubClass(SubSubClass):
        pass

    classes = get_flattened_subclasses(BaseClass)
    assert len(classes) == 3
    assert SubClass in classes
    assert SubSubClass in classes
    assert SubSubSubClass in classes



from collections import namedtuple

import pytest

from leapp.reporting import _add_to_dict, create_report_from_error


ReportPrimitive = namedtuple('ReportPrimitive', ['data', 'path', 'value', 'is_leaf_list'])

leaf_is_dict = ReportPrimitive({}, ('summary',), 'We better fix this!', False)
leaf_is_list = ReportPrimitive({}, ('summary',), 'We better fix this!', False)

path_taken = ReportPrimitive({'summary': 'Fix is coming!'}, ('summary',), 'We better fix this!', False)

leaf_list_append1 = ReportPrimitive(
    {}, ('detail', 'remediation'), {'type': 'hint', 'context': 'Read the docs!'}, True
)
leaf_list_append2 = ReportPrimitive(
    {}, ('detail', 'remediation'), {'type': 'command', 'context': 'man leapp'}, True
)


def value_from_path(data, path):
    value = data
    for p in path:
        if isinstance(value, dict) and value.get(p, None):
            value = value[p]
        else:
            return None
    return value


def is_path_valid(data, path):
    return bool(value_from_path(data, path))


def assert_leaf_list(data, path, is_leaf_list):
    res = isinstance(value_from_path(data, path), list)
    if is_leaf_list:
        assert res
    else:
        assert not res


@pytest.mark.parametrize('primitive', (leaf_is_dict, leaf_is_list))
def test_add_to_dict_func_leaves(primitive):
    data, path, value, is_leaf_list = primitive

    _add_to_dict(data, path, value, is_leaf_list)
    assert is_path_valid(data, path)
    assert_leaf_list(data, path, is_leaf_list)


def test_add_to_dict_func_path_taken():
    data, path, value, is_leaf_list = path_taken

    with pytest.raises(ValueError):
        _add_to_dict(data, path, value, is_leaf_list)


def test_add_to_dict_func_append():
    data, path, value, is_leaf_list = leaf_list_append1

    _add_to_dict(data, path, value, is_leaf_list)
    assert is_path_valid(data, path)
    assert_leaf_list(data, path, is_leaf_list)

    # we do not want data again as we will re-use the dict for appending
    _data, path, value, is_leaf_list = leaf_list_append2
    _add_to_dict(data, path, value, is_leaf_list)

    assert is_path_valid(data, path)
    assert_leaf_list(data, path, is_leaf_list)
    assert len(value_from_path(data, path)) == 2


def test_convert_from_error_to_report():
    error_dict_no_details = {
        'message': 'The system is not registered or subscribed.',
        'time': '2019-11-19T05:13:04.447562Z',
        'details': None,
        'actor': 'verify_check_results',
        'severity': 'error'}
    report = create_report_from_error(error_dict_no_details)
    assert report["severity"] == "high"
    assert report["title"] == "The system is not registered or subscribed."
    assert report["audience"] == "sysadmin"
    assert report["summary"] == ""
    error_dict_with_details = {
        'message': 'The system is not registered or subscribed.',
        'time': '2019-11-19T05:13:04.447562Z',
        'details': 'Some other info that should go to report summary',
        'actor': 'verify_check_results',
        'severity': 'error'}
    report = create_report_from_error(error_dict_with_details)
    assert report["severity"] == "high"
    assert report["title"] == "The system is not registered or subscribed."
    assert report["audience"] == "sysadmin"
    assert report["summary"] == "Some other info that should go to report summary"



import os
from subprocess import check_call

import pytest

TESTING_REPOSITORY_NAME = 'testing'


def make_repository_dir_fixture(name, repository_name=TESTING_REPOSITORY_NAME, scope='session'):
    @pytest.fixture(scope=scope, name=name)
    def impl(request, tmpdir_factory):
        old_value = os.environ.get('LEAPP_CONFIG', None)

        if old_value is not None:
            def fin():
                os.environ['LEAPP_CONFIG'] = old_value
            request.addfinalizer(fin)

        root = tmpdir_factory.mktemp('repositories')
        with root.as_cwd():
            check_call(['snactor', 'repo', 'new', repository_name])
            repository = root.join(repository_name)
            os.environ['LEAPP_CONFIG'] = repository.join('.leapp', 'leapp.conf').strpath
            return repository
    return impl



"""
:py:mod:`leapp.libraries.actor` represents the import location for private actor libraries that
are placed in the actor's libraries folder.


Example:
      If you actor has a libraries folder with a module.py python module, import it
      from the actor like this::

      from leapp.libraries.actor import module

"""



import logging
import os
import sys

from leapp.compat import string_types
from leapp.dialogs import Dialog
from leapp.exceptions import (MissingActorAttributeError, RequestStopAfterPhase, StopActorExecution,
                              StopActorExecutionError, WorkflowConfigNotAvailable, WrongAttributeTypeError)
from leapp.models import DialogModel, Model
from leapp.models.error_severity import ErrorSeverity
from leapp.tags import Tag
from leapp.utils import get_api_models
from leapp.utils.i18n import install_translation_for_actor
from leapp.utils.meta import get_flattened_subclasses
from leapp.workflows.api import WorkflowAPI


class Actor(object):
    """
    The Actor class represents the smallest step in the workflow. It defines what kind
    of data it expects, it consumes (processes) the given data, and it produces data for other
    actors in the workflow.
    """

    current_instance = None
    """
    Instance of the currently executed actor. Within a process only exist one Actor instance. This will allow
    convenience functions for library developers to be available.
    """

    ErrorSeverity = ErrorSeverity
    """ Convenience forward for the :py:class:`leapp.models.error_severity.ErrorSeverity` constants. """

    name = None
    """ Name of the actor that is used to identify data or messages created by the actor. """

    description = None
    """
    .. deprecated:: 0.5.0
       Write the actor's description as a docstring.
    """

    consumes = ()
    """
    Tuple of :py:class:`leapp.models.Model` derived classes defined in the :ref:`repositories <terminology:repository>`
    that define :ref:`messages <terminology:message>` the actor consumes.
    """

    produces = ()
    """
    Tuple of :py:class:`leapp.models.Model` derived classes defined in the :ref:`repositories <terminology:repository>`
    that define :ref:`messages <terminology:message>` the actor produces.
    """

    tags = ()
    """
    Tuple of :py:class:`leapp.tags.Tag` derived classes by which :ref:`workflow <terminology:workflow>`
    :ref:`phases <terminology:phase>` select actors for execution.
    """

    dialogs = ()
    """
    Tuple of :py:class:`leapp.dialogs.dialog.Dialog` derived classes that define questions to ask the user.
    Dialogs that are added to this list allow for persisting answers the user has given in the answer file storage.
    """

    apis = ()
    """
    Tuple of :py:class:`leapp.workflow.api.WorkflowAPI` derived classes that implement Workflow APIs that are used by
    an actor. Any models the apis produce or consume will be considered by the framework as if the actor defined them.
    """

    text_domain = None
    """
    Using text domain allows to override the default gettext text domain, for custom localization support.
    The default locale installation location is used which usually is /usr/share/locale
    """

    def serialize(self):
        """
        :return: Serialized information for the actor
        """
        return {
            'name': self.name,
            'path': os.path.dirname(sys.modules[type(self).__module__].__file__),
            'class_name': type(self).__name__,
            'description': self.description or type(self).__doc__,
            'consumes': [c.__name__ for c in self.consumes],
            'produces': [p.__name__ for p in self.produces],
            'tags': [t.__name__ for t in self.tags],
            'dialogs': [d.serialize() for d in self.dialogs],
        }

    def __init__(self, messaging=None, logger=None, config_model=None, skip_dialogs=False):
        self._configuration = None
        """
        Instance a workflow defined configuration model if available.

        This depends on the definition of such a configuration model being defined by the workflow
        and an actor that provides such a message.
        """
        Actor.current_instance = self
        install_translation_for_actor(type(self))
        self._messaging = messaging
        self.log = (logger or logging.getLogger('leapp.actors')).getChild(self.name)
        self.skip_dialogs = skip_dialogs
        """ A configured logger instance for the current actor. """

        if config_model:
            self._configuration = next(self.consume(config_model), None)

        # Needed so produce allows to send messages for models specified also by workflow APIs
        type(self).produces = get_api_models(type(self), 'produces')

    def get_answers(self, dialog):
        """
        Gets the answers for a dialog. The dialog needs be predefined in :py:attr:`dialogs`.

        :param dialog: Dialog instance to show
        :return: dictionary with the requested answers, None if not a defined dialog
        """
        self._messaging.register_dialog(dialog, self)
        if dialog in type(self).dialogs:
            if self.skip_dialogs:
                # non-interactive mode of operation
                return self._messaging.get_answers(dialog)
            return self._messaging.request_answers(dialog)
        return None

    def show_message(self, message):
        """
        Display a message in user interterface currently in use (CLI, GUI).

        It uses one of the dialog renderers in :py:mod:`leapp.dialogs.renderer`.

        :param message: Message to show
        :type message: str
        """
        self._messaging.show_message(message)

    @property
    def configuration(self):
        """
        Returns the config model generated by specific workflow configuration actor
        """
        if not self._configuration:
            raise WorkflowConfigNotAvailable(self.name)
        return self._configuration

    @property
    def actor_files_paths(self):
        """
        Returns the file paths that are bundled with the actor. (Path to the content of the actor's file directory).
        """
        return os.getenv("LEAPP_FILES", "").split(":")

    @property
    def files_paths(self):
        """ Returns all actor file paths related to the actor and common actors file paths. """
        return self.actor_files_paths + self.common_files_paths

    @property
    def common_files_paths(self):
        """ Returns all common repository file paths. """
        return os.getenv("LEAPP_COMMON_FILES", "").split(":")

    @property
    def actor_tools_paths(self):
        """
        Returns the tool paths that are bundled with the actor. (Path to the content of the actor's tools directory).
        """
        return os.getenv("LEAPP_TOOLS", "").split(":")

    @property
    def common_tools_paths(self):
        """ Returns all common repository tool paths. """
        return os.getenv("LEAPP_COMMON_TOOLS", "").split(":")

    @property
    def tools_paths(self):
        """ Returns all actor tools paths related to the actor and common actors tools paths. """
        return self.actor_tools_paths + self.common_tools_paths

    @staticmethod
    def _get_folder_path(directories, name):
        """
        Finds the first matching folder path within directories.

        :param name: Name of the folder
        :type name: str
        :return: Found folder path
        :rtype: str or None
        """
        for path in directories:
            path = os.path.join(path, name)
            if os.path.isdir(path):
                return path
        return None

    def get_folder_path(self, name):
        """
        Finds the first matching folder path within :py:attr:`files_paths`.

        :param name: Name of the folder
        :type name: str
        :return: Found folder path
        :rtype: str or None
        """
        return self._get_folder_path(self.files_paths, name)

    def get_common_folder_path(self, name):
        """
        Finds the first matching folder path within :py:attr:`common_files_paths`.

        :param name: Name of the folder
        :type name: str
        :return: Found folder path
        :rtype: str or None
        """
        return self._get_folder_path(self.common_files_paths, name)

    def get_actor_folder_path(self, name):
        """
        Finds the first matching folder path within :py:attr:`actor_files_paths`.

        :param name: Name of the folder
        :type name: str
        :return: Found folder path
        :rtype: str or None
        """
        return self._get_folder_path(self.actor_files_paths, name)

    @staticmethod
    def _get_file_path(directories, name):
        """
        Finds the first matching file path within directories.

        :param name: Name of the file
        :type name: str
        :return: Found file path
        :rtype: str or None
        """
        for path in directories:
            path = os.path.join(path, name)
            if os.path.isfile(path):
                return path
        return None

    def get_file_path(self, name):
        """
        Finds the first matching file path within :py:attr:`files_paths`.

        :param name: Name of the file
        :type name: str
        :return: Found file path
        :rtype: str or None
        """
        return self._get_file_path(self.files_paths, name)

    def get_common_file_path(self, name):
        """
        Finds the first matching file path within :py:attr:`common_files_paths`.

        :param name: Name of the file
        :type name: str
        :return: Found file path
        :rtype: str or None
        """
        return self._get_file_path(self.common_files_paths, name)

    def get_actor_file_path(self, name):
        """
        Finds the first matching file path within :py:attr:`actor_files_paths`.

        :param name: Name of the file
        :type name: str
        :return: Found file path
        :rtype: str or None
        """
        return self._get_file_path(self.actor_files_paths, name)

    @staticmethod
    def _get_tool_path(directories, name):
        """
        Finds the first matching executable file within directories.

        :param name: Name of the file
        :type name: str
        :return: Found file path
        :rtype: str or None
        """
        for path in directories:
            path = os.path.join(path, name)
            if os.path.isfile(path) and os.access(path, os.X_OK):
                return path
        return None

    def get_tool_path(self, name):
        """
        Finds the first matching executable file path within :py:attr:`tools_paths`.

        :param name: Name of the file
        :type name: str
        :return: Found file path
        :rtype: str or None
        """
        return self._get_tool_path(self.tools_paths, name)

    def get_common_tool_path(self, name):
        """
        Finds the first matching executable file path within :py:attr:`common_tools_paths`.

        :param name: Name of the file
        :type name: str
        :return: Found file path
        :rtype: str or None
        """
        return self._get_tool_path(self.common_tools_paths, name)

    def get_actor_tool_path(self, name):
        """
        Finds the first matching executable file path within :py:attr:`actor_tools_paths`.

        :param name: Name of the file
        :type name: str
        :return: Found file path
        :rtype: str or None
        """
        return self._get_tool_path(self.actor_tools_paths, name)

    def run(self, *args):
        """ Runs the actor calling the method :py:func:`process`. """
        os.environ['LEAPP_CURRENT_ACTOR'] = self.name
        try:
            self.process(*args)
        except StopActorExecution:
            pass
        except StopActorExecutionError as err:
            self.report_error(err.message, err.severity, err.details)
        except RequestStopAfterPhase:
            self._messaging.request_stop_after_phase()
        finally:
            os.environ.pop('LEAPP_CURRENT_ACTOR', None)

    def process(self, *args, **kwargs):
        """ Main processing method. In inherited actors, the function needs to be defined to be able to be processed."""
        raise NotImplementedError()

    def produce(self, *models):
        """
        By calling produce, model instances are stored as messages. Those messages can be then consumed by other actors.

        :param models: Messages to be sent (those model types have to be specified in :py:attr:`produces`
        :type models: Variable number of the derived classes from :py:class:`leapp.models.Model`
        """
        if self._messaging:
            for model in models:
                if isinstance(model, type(self).produces):
                    self._messaging.produce(model, self)
                else:
                    self.log.warning('Actor is trying to produce a message of type "{}" without mentioning it '
                                     'explicitely in the actor\'s "produces" tuple. The message will be ignored'.format(
                                         type(model)))

    def consume(self, *models):
        """
        Retrieve messages specified in the actors :py:attr:`consumes` attribute, and filter message types by
        models.

        :param models: Models to use as a filter for the messages to return
        :type models: Variable number of the derived classes from :py:class:`leapp.models.Model`
        :return: All messages of the specified model(s) produced by other actors
        :rtype: Iterable with messages or an empty tuple
        """
        if self._messaging:
            return self._messaging.consume(self, *models)
        return ()

    def report_error(self, message, severity=ErrorSeverity.ERROR, details=None):
        """
        Reports an execution error

        :param message: A message to print the possible error
        :type message: str
        :param severity: Severity of the error default :py:attr:`leapp.messaging.errors.ErrorSeverity.ERROR`
        :type severity: str with defined values from :py:attr:`leapp.messaging.errors.ErrorSeverity.ERROR`
        :param details: A dictionary where additional context information is passed along with the error
        :type details: dict
        :return: None
        """
        if self._messaging:
            if not ErrorSeverity.validate(severity):
                self.log.warning("report_error: Unknown severity value %s was passed - Falling back to ERROR", severity)
                severity = ErrorSeverity.ERROR
            self._messaging.report_error(
                message=message,
                severity=severity,
                actor=self,
                details=details)


def _is_type(value_type):
    def validate(actor, name, value):
        if not isinstance(value, value_type):
            raise WrongAttributeTypeError('Actor {} attribute {} should be of the type {}'.format(actor, name,
                                                                                                  value_type))
        return value
    return validate


def _is_tuple_of(value_type):
    def validate(actor, name, value):
        _is_type(tuple)(actor, name, value)
        if not value:
            raise WrongAttributeTypeError(
                'Actor {} attribute {} should contain at least one item of the type {}'.format(actor, name, value_type))
        if not all([isinstance(item, value_type) for item in value]):
            raise WrongAttributeTypeError(
                'Actor {} attribute {} should contain only values of the type {}'.format(actor, name, value_type))
        return value
    return validate


def _lint_warn(actor, name, type_name):
    warnings = getattr(actor, '_warnings', {})
    if not warnings.get(name + '_tuple'):
        warnings[name + '_tuple'] = True
        setattr(actor, '_warnings', warnings)
        logging.getLogger("leapp.linter").warning("Actor %s field %s should be a tuple of %s", actor, name, type_name)


def _is_model_tuple(actor, name, value):
    if isinstance(value, type) and issubclass(value, Model):
        _lint_warn(actor, name, "Models")
        value = (value,)
    _is_type(tuple)(actor, name, value)
    if not all([True] + [isinstance(item, type) and issubclass(item, Model) for item in value]):
        raise WrongAttributeTypeError(
            'Actor {} attribute {} should contain only Models'.format(actor, name))
    return value


def _is_dialog_tuple(actor, name, value):
    if isinstance(value, Dialog):
        _lint_warn(actor, name, "Dialogs")
        value = (value,)
    _is_type(tuple)(actor, name, value)
    if not all([True] + [isinstance(item, Dialog) for item in value]):
        raise WrongAttributeTypeError(
            'Actor {} attribute {} should contain only Dialogs'.format(actor, name))
    return value


def _is_tag_tuple(actor, name, value):
    if isinstance(value, type) and issubclass(value, Tag):
        _lint_warn(actor, name, "Tags")
        value = (value,)
    _is_type(tuple)(actor, name, value)
    if not all([True] + [isinstance(item, type) and issubclass(item, Tag) for item in value]):
        raise WrongAttributeTypeError(
            'Actor {} attribute {} should contain only Tags'.format(actor, name))
    return value


def _is_api_tuple(actor, name, value):
    if isinstance(value, type) and issubclass(value, WorkflowAPI):
        _lint_warn(actor, name, "Apis")
        value = (value,)
    _is_type(tuple)(actor, name, value)
    if not all([True] + [isinstance(item, type) and issubclass(item, WorkflowAPI) for item in value]):
        raise WrongAttributeTypeError(
            'Actor {} attribute {} should contain only WorkflowAPIs'.format(actor, name))
    return value


def _get_attribute(actor, name, validator, required=False, default_value=None, additional_info='', resolve=None):
    if resolve:
        value = resolve(actor, name)
    else:
        value = getattr(actor, name, None)
    if not value and required:
        raise MissingActorAttributeError('Actor {} is missing attribute {}.{}'.format(actor, name, additional_info))
    if value or required:
        value = validator(actor, name, value)
    if not value and default_value is not None:
        value = default_value
    return name, value


def get_actor_metadata(actor):
    """
    Creates Actor's metadata dictionary

    :param actor: Actor whose metadata are needed
    :type actor: derived class from :py:class:`leapp.actors.Actor`
    :return: Dictionary with the name, tags, consumes, produces, and description of the actor
    """
    additional_tag_info = ' At least one tag is required for actors. Please fill the tags field'
    return dict([
        ('class_name', actor.__name__),
        # TODO: missing tests. We need to ensure this doesn't break anything.
        # # OTOH, actor_definition.inspect_actor ends with empty list on Python3
        # # if path is not transformed into the realpath.
        ('path', os.path.dirname(os.path.realpath(sys.modules[actor.__module__].__file__))),
        _get_attribute(actor, 'name', _is_type(string_types), required=True),
        _get_attribute(actor, 'tags', _is_tag_tuple, required=True, additional_info=additional_tag_info),
        _get_attribute(actor, 'consumes', _is_model_tuple, required=False, default_value=(), resolve=get_api_models),
        _get_attribute(actor, 'produces', _is_model_tuple, required=False, default_value=(), resolve=get_api_models),
        _get_attribute(actor, 'dialogs', _is_dialog_tuple, required=False, default_value=()),
        _get_attribute(actor, 'description', _is_type(string_types), required=False,
                       default_value=actor.__doc__ or 'There has been no description provided for this actor.'),
        _get_attribute(actor, 'apis', _is_api_tuple, required=False, default_value=())
    ])


def get_actors():
    """
    :return: All registered actors with their metadata
    """
    actors = get_flattened_subclasses(Actor)
    for actor in actors:
        get_actor_metadata(actor)
    return actors



from __future__ import print_function

import errno
import platform
import select
from collections import defaultdict

POLL_NULL = 0
POLL_IN = 1
POLL_PRI = 2
POLL_OUT = 4
POLL_ERR = 8
POLL_HUP = 16


class EventLoopKQUEUE(object):
    MAX_EVENTS = 1024

    def __init__(self):
        self._kqueue = select.kqueue()
        self._fds = {}

    def _control(self, fd, mode, flags):
        events = []
        if mode & POLL_OUT:
            events.append(select.kevent(fd, select.KQ_FILTER_WRITE, flags))
        if mode & POLL_IN:
            events.append(select.kevent(fd, select.KQ_FILTER_READ, flags))
        for e in events:
            try:
                self._kqueue.control([e], 0)
            except OSError as exc:
                if exc.errno not in (errno.EBADF, errno.ENOENT):
                    raise

    def poll(self, timeout):
        if timeout < 0:
            timeout = None  # kqueue behaviour
        events = self._kqueue.control(None, EventLoopKQUEUE.MAX_EVENTS, timeout)
        results = defaultdict(lambda: POLL_NULL)
        for e in events:
            fd = e.ident
            print(e)
            if e.flags & select.KQ_EV_EOF and e.data == 0:
                results[fd] = POLL_HUP
            elif e.filter == select.KQ_FILTER_READ:
                results[fd] |= POLL_IN
            elif e.filter == select.KQ_FILTER_WRITE:
                results[fd] |= POLL_OUT
        return results.items()

    def register(self, fd, mode):
        self._fds[fd] = mode
        self._control(fd, mode, select.KQ_EV_ADD)

    def unregister(self, fd):
        self._control(fd, self._fds[fd], select.KQ_EV_DELETE)
        del self._fds[fd]

    def modify(self, fd, mode):
        self.unregister(fd)
        self.register(fd, mode)

    def close(self):
        self._kqueue.close()

    @property
    def closed(self):
        return self._kqueue.closed


EventLoop = EventLoopKQUEUE if platform.system() == 'Darwin' else select.epoll



import pytest

from leapp.actors import Actor, get_actors
from leapp.topics import Topic
from leapp.exceptions import CyclingDependenciesError
from leapp.models import Model
from leapp.tags import Tag
from leapp.workflows.phaseactors import PhaseActors


class CycleTag1(Tag):
    name = 'cycle1'


class PhaseActorsModelsTag1(Tag):
    name = 'phase-actor-models1'


class CycleTopic(Topic):
    name = 'cycle'


class CycleModel1(Model):
    topic = CycleTopic


class CycleModel2(Model):
    topic = CycleTopic


class CycleActor1(Actor):
    name = 'CycleActor1'
    description = 'Unit Test Actor CycleActor1'
    consumes = (CycleModel1,)
    produces = (CycleModel2,)
    tags = (CycleTag1,)

    def process(self):
        pass


class CycleActor2(Actor):
    name = 'CycleActor2'
    description = 'Unit Test Actor CycleActor2'
    consumes = (CycleModel2,)
    produces = (CycleModel1,)
    tags = (CycleTag1, PhaseActorsModelsTag1)

    def process(self):
        pass


class CycleActor3(Actor):
    name = 'CycleActor3'
    description = 'Unit Test Actor CycleActor3'
    consumes = (CycleModel1,)
    produces = ()
    tags = (CycleTag1, PhaseActorsModelsTag1)

    def process(self):
        pass


def setup_module():
    for actor in get_actors():
        for tag in actor.tags:
            tag.actors = tag.actors + (actor,)


def test_actor_phases_detect_cycles():
    # Expected a cycle to be detected
    with pytest.raises(CyclingDependenciesError):
        PhaseActors(CycleTag1.actors, 'Test')

    # This should not cause a cycle to be present
    PhaseActors(PhaseActorsModelsTag1.actors, 'Test')


def test_actor_phases_check_models():
    phase_actors = PhaseActors(PhaseActorsModelsTag1.actors, 'Test')
    assert len(phase_actors.initial) == 1 and phase_actors.initial[0] is CycleModel2

    assert len(phase_actors.consumes) == 2
    assert CycleModel1 in phase_actors.consumes
    assert CycleModel2 in phase_actors.consumes

    assert len(phase_actors.produces) == 1
    assert CycleModel1 in phase_actors.produces


def test_actor_phases_order():
    initial_actors = (CycleActor3, CycleActor2)
    phase_actors = PhaseActors(initial_actors, 'Test')

    assert len(phase_actors.actors) == 2
    assert phase_actors.actors[0] is CycleActor2
    assert phase_actors.actors[1] is CycleActor3



class ErrorSeverity(object):
    FATAL = 'fatal'
    ERROR = 'error'
    WARNING = 'warning'

    ALLOWED_VALUES = (FATAL, ERROR, WARNING)

    @classmethod
    def validate(cls, value):
        return value in cls.ALLOWED_VALUES



from leapp.actors import Actor
from leapp.tags import ActorFileApiTag
from leapp.models import ApiTestConsume, ApiTestProduce


class Second(Actor):
    name = 'second'
    description = 'No description has been provided for the second actor.'
    consumes = (ApiTestConsume,)
    produces = (ApiTestProduce,)
    tags = (ActorFileApiTag,)

    def process(self):
        pass



import datetime
import json

from leapp.compat import string_types
from leapp.models import fields, Model, ErrorModel
from leapp.topics import ReportTopic
from leapp.libraries.stdlib.api import produce


class Report(Model):
    """
    Framework model used for reporting
    """

    __non_inheritable__ = True

    topic = ReportTopic

    report = fields.JSON()


def _add_to_dict(data, path, value, leaf_list=False):
    for p in path[:-1]:
        if p not in data:
            data[p] = {}
        data = data[p]

    if leaf_list:
        data[path[-1]] = data.get(path[-1], []) + [value]
    else:
        if path[-1] in data:
            raise ValueError('Path {} is already taken'.format('.'.join(path)))
        data[path[-1]] = value


class BasePrimitive(object):
    """
    Report primitive (field) base class for dict targets (implies unique path)
    """
    # deepest nested key in resulting report dict
    name = ''

    def __init__(self, value=None):
        if not isinstance(value, string_types):
            raise TypeError('Value of "{}" must be a string'.format(self.__class__.__name__))
        self._value = value

    @property
    def value(self):
        """Return report field value"""
        return self._value

    @property
    def path(self):
        """Return report field destination path (nested dict description)"""
        return (self.name, )

    def to_dict(self):
        return {self.name: self._value}

    def apply(self, report):
        """Add report entry to the report dict based on provided path"""
        _add_to_dict(report, self.path, self.value, leaf_list=False)


class BaseListPrimitive(BasePrimitive):
    """
    Report primitive (field) base class for list targets (we append value to a list)
    """

    def apply(self, report):
        """Add report entry to the report dict based on provided path"""
        _add_to_dict(report, self.path, self.value, leaf_list=True)


class Title(BasePrimitive):
    """Report title"""
    name = 'title'


class Summary(BasePrimitive):
    """Report summary"""
    name = 'summary'


class Severity(BasePrimitive):
    """Report severity"""
    name = 'severity'

    INFO = 'info'
    LOW = 'low'
    MEDIUM = 'medium'
    HIGH = 'high'

    def __init__(self, value=None):
        severities = (self.INFO, self.LOW, self.MEDIUM, self.HIGH)
        if value not in severities:
            raise ValueError('Severity should be one of: "{}"'.format(' '.join(severities)))
        self._value = value


class Audience(BasePrimitive):
    """Target audience of the report"""
    name = 'audience'

    def __init__(self, value=None):
        if not isinstance(value, string_types):
            raise TypeError('Value of "Audience" must be a string')
        audiences = ('sysadmin', 'developer')
        if value not in audiences:
            raise ValueError('Audience should be one of: "{}"'.format(' '.join(audiences)))
        self._value = value


class Flags(BasePrimitive):
    """Report flags"""
    name = 'flags'

    INHIBITOR = 'inhibitor'
    FAILURE = 'failure'

    def __init__(self, value=None):
        if not isinstance(value, list):
            raise TypeError('Value of "Flags" must be a list')
        self._value = value


class Tags(BasePrimitive):
    """Report tags"""
    name = 'tags'

    class _Value(object):
        def __init__(self, value):
            self.value = value

    ACCESSIBILITY = _Value('accessibility')
    AUTHENTICATION = _Value('authentication')
    BOOT = _Value('boot')
    COMMUNICATION = _Value('communication')
    DESKTOP = _Value('desktop environment')
    DRIVERS = _Value('drivers')
    EMAIL = _Value('email')
    ENCRYPTION = _Value('encryption')
    FILESYSTEM = _Value('filesystem')
    FIREWALL = _Value('firewall')
    HIGH_AVAILABILITY = _Value('high availability')
    KERNEL = _Value('kernel')
    MONITORING = _Value('monitoring')
    NETWORK = _Value('network')
    OS_FACTS = _Value('OS facts')
    PUBLIC_CLOUD = _Value('public cloud')
    PYTHON = _Value('python')
    REPOSITORY = _Value('repository')
    RHUI = _Value('rhui')
    SANITY = _Value('sanity')
    SECURITY = _Value('security')
    SELINUX = _Value('selinux')
    SERVICES = _Value('services')
    TIME_MANAGEMENT = _Value('time management')
    TOOLS = _Value('tools')
    UPGRADE_PROCESS = _Value('upgrade process')

    def __init__(self, value=None):
        if not isinstance(value, list):
            raise TypeError('Value of "Tags" must be a list')
        if not all([isinstance(v, Tags._Value) for v in value]):
            raise TypeError('Unsupported tag value passed for Report Tags.')
        # after the objects validation we need the actual values in the list
        self._value = [v.value for v in value]


class ExternalLink(BaseListPrimitive):
    """External link report detail field"""
    name = 'external'

    def __init__(self, url=None, title=None):
        if not all(isinstance(v, string_types) for v in (url, title)):
            raise TypeError('Values "url" and "title" of "ExternalLink" must be a string')
        self._value = {'url': url, 'title': title}

    @property
    def path(self):
        return ('detail', 'external')


class RelatedResource(BaseListPrimitive):
    """Report detail field for related resources (e.g. affected packages/files)"""
    name = 'related_resources'

    def __init__(self, scheme=None, identifier=None):
        if not all(isinstance(v, string_types) for v in (scheme, identifier)):
            raise TypeError('Values "scheme" and "identifier" of "RelatedResource" must be a string')
        self._value = {'scheme': scheme, 'title': identifier}

    @property
    def path(self):
        return ('detail', 'related_resources')


class BaseRemediation(BaseListPrimitive):
    """Remediation base class"""
    name = 'remediations'

    @property
    def path(self):
        return ('detail', 'remediations')


class RemediationCommand(BaseRemediation):
    def __init__(self, value=None):
        if not isinstance(value, list):
            raise TypeError('Value of "RemediationCommand" must be a list')
        self._value = {'type': 'command', 'context': value}

    def __repr__(self):
        return "[{}] {}".format(self._value['type'], ' '.join(self._value['context']))


class RemediationHint(BaseRemediation):
    def __init__(self, value=None):
        self._value = {'type': 'hint', 'context': value}

    def __repr__(self):
        return "[{}] {}".format(self._value['type'], self._value['context'])


class RemediationPlaybook(BaseRemediation):
    def __init__(self, value=None):
        self._value = {'type': 'playbook', 'context': value}

    def __repr__(self):
        return "[{}] {}".format(self._value['type'], self._value['context'])


class Remediation(object):
    def __init__(self, playbook=None, commands=None, hint=None):
        self._remediations = []
        if hint:
            self._remediations.append(RemediationHint(value=hint))
        if commands:
            for command in commands:
                self._remediations.append(RemediationCommand(value=command))
        if playbook:
            self._remediations.append(RemediationPlaybook(value=playbook))

    def apply(self, report):
        for remediation in self._remediations:
            remediation.apply(report)

    def __repr__(self):
        return "\n".join([repr(r) for r in self._remediations])

    def to_dict(self):
        return {'remediations': [r.to_dict() for r in self._remediations]}

    @classmethod
    def from_dict(cls, data):
        values = data.get('remediations', [])
        playbook = next((v['context'] for v in values if v['type'] == 'playbook'), None)
        hint = next((v['context'] for v in values if v['type'] == 'hint'), None)
        commands = [v['context'] for v in values if v['type'] == 'command']
        if values:
            return Remediation(playbook, commands, hint)


def _sanitize_entries(entries):
    if not any(isinstance(e, Title) for e in entries):
        raise ValueError('Title not provided in the report.')
    if not any(isinstance(e, Summary) for e in entries):
        raise ValueError('Summary not provided in the report.')
    if not any(isinstance(e, Severity) for e in entries):
        entries.append(Severity('info'))
    if not any(isinstance(e, Audience) for e in entries):
        entries.append(Audience('sysadmin'))


def _create_report_object(entries):
    report = {}

    _sanitize_entries(entries)
    for entry in entries:
        entry.apply(report)

    return Report(report=report)


def create_report(entries):
    """
    Produce final report message
    """
    produce(_create_report_object(entries))


_DEPRECATION_SEVERITY_THRESHOLD = datetime.timedelta(days=180)


def create_report_from_deprecation(deprecation):
    """
    Convert deprecation json to report json
    """
    since = datetime.datetime.strptime(deprecation['since'], '%Y-%m-%d').date()

    severity = Severity.MEDIUM
    if since < (datetime.date.today() - _DEPRECATION_SEVERITY_THRESHOLD):
        severity = Severity.HIGH

    report = _create_report_object([
        Title('{message} at {filename}:{lineno}'.format(
            message=deprecation['message'].split('\n')[0],
            filename=deprecation['filename'],
            lineno=deprecation['lineno'])),
        Severity(severity),
        Audience('developer'),
        Summary('{reason}\nSince: {since}\nLocation: {filename}:{lineno}\nNear: {line}'.format(
            reason=deprecation['reason'],
            since=deprecation['since'],
            filename=deprecation['filename'],
            lineno=deprecation['lineno'],
            line=deprecation['line']
        )),
    ])
    return json.loads(report.dump()['report'])


def create_report_from_error(error_dict):
    """
    Convert error json to report json
    """
    error = ErrorModel.create(error_dict)
    entries = [Title(error.message), Summary(error.details or ""), Severity('high'), Audience('sysadmin')]
    report = _create_report_object(entries)
    return json.loads(report.dump()['report'])



import json
import os
from subprocess import check_call, check_output, CalledProcessError, STDOUT

import pytest

from leapp.snactor import context
from leapp.utils import audit


def setup_module():
    os.environ['PYTHONDONTWRITEBYTECODE'] = '1'


def test_discovery(repository_dir):
    with repository_dir.as_cwd():
        check_call(['snactor', 'discover'])

        # Ensure snactor discover --json returns valid json
        output = check_output(['snactor', 'discover', '--json']).decode('utf-8')
        data = json.loads(output)
        assert 'actors' in data
        assert 'base_dir' in data and repository_dir.samefile(data['base_dir'])
        assert 'models' in data
        assert 'repository' in data
        assert 'tags' in data
        assert 'topics' in data

    with type(repository_dir)(path=repository_dir.dirname).as_cwd():
        with pytest.raises(CalledProcessError):
            check_call(['snactor', 'discover'])


def test_raises_create_outside_repository():
    for cmd in ([['snactor', 'new-tag', 'winteriscoming'],
                 ['snactor', 'new-topic', 'winteriscoming'],
                 ['snactor', 'new-model', 'winteriscoming', '--topic', 'WinteriscomingTopic']]):
        with pytest.raises(CalledProcessError) as err:
            check_output(cmd, stderr=STDOUT)
        assert ('This command must be executed from the repository directory' in
                err.value.output.decode('utf-8'))


def test_new_tag(repository_dir):
    with repository_dir.as_cwd():
        check_call(['snactor', 'new-tag', 'Test'])
        assert repository_dir.join('tags/test.py').check(file=True)
        check_call(['snactor', 'discover'])


def test_new_topic(repository_dir):
    with repository_dir.as_cwd():
        # We need the topic to be created already for the model
        # So we have to check if it wasn't already created
        if not repository_dir.join('topics/test.py').check(file=True):
            check_call(['snactor', 'new-topic', 'Test'])
        assert repository_dir.join('topics/test.py').check(file=True)
        check_call(['snactor', 'discover'])


def test_different_objects_same_name_discover(repository_dir):
    with repository_dir.as_cwd():
        check_call(['snactor', 'new-tag', 'winteriscoming'])
        check_call(['snactor', 'new-topic', 'winteriscoming'])
        check_call(['snactor', 'new-model', 'winteriscoming', '--topic', 'WinteriscomingTopic'])
        objs = ['tags/winteriscoming.py', 'topics/winteriscoming.py', 'models/winteriscoming.py']
        for obj in objs:
            assert repository_dir.join(obj).check(file=True)
        out = check_output(['snactor', 'discover'])
        for obj in objs:
            assert out.find(obj.encode('utf-8')) > -1


def test_new_model(repository_dir):
    # We need the topic to be created already
    if not repository_dir.join('topics/test.py').check(file=True):
        test_new_topic(repository_dir)
    with repository_dir.as_cwd():
        # We need the model to be created already for the actor
        # So we have to check if it wasn't already created
        if not repository_dir.join('models/testmodel.py').check(file=True):
            check_call(['snactor', 'new-model', 'TestModel'])
        assert repository_dir.join('models/testmodel.py').check(file=True)
        with pytest.raises(CalledProcessError):
            # Now discover should fail due to the missing topic
            check_call(['snactor', 'discover'])
        repository_dir.join('models/testmodel.py').write('''
from leapp.models import Model, fields
from leapp.topics import TestTopic


class TestModel(Model):
    topic = TestTopic
    value = fields.String(default='Test value')
''')
        check_call(['snactor', 'discover'])


def test_ref_model(repository_dir):
    # We need the topic to be created already
    if not repository_dir.join('topics/test.py').check(file=True):
        test_new_topic(repository_dir)
    with repository_dir.as_cwd():
        # We need the model to be created already for the actor
        # So we have to check if it wasn't already created
        if not repository_dir.join('models/a.py').check(file=True):
            check_call(['snactor', 'new-model', 'A'])
        assert repository_dir.join('models/a.py').check(file=True)
        with pytest.raises(CalledProcessError):
            # Now discover should fail due to the missing topic
            check_call(['snactor', 'discover'])
        repository_dir.join('models/a.py').write('''
from leapp.models import Model, fields, TestModel
from leapp.topics import TestTopic


class A(Model):
    topic = TestTopic
    referenced = fields.Model(TestModel)
''')
        check_call(['snactor', 'discover'])


def test_new_actor(repository_dir):
    # We need the model to be created already
    if not repository_dir.join('models/testmodel.py').check(file=True):
        test_new_model(repository_dir)
    with repository_dir.as_cwd():
        check_call(['snactor', 'new-actor', 'Test'])
        assert repository_dir.join('actors/test/actor.py').check(file=True)
        with pytest.raises(CalledProcessError):
            check_call(['snactor', 'discover'])
        repository_dir.join('actors/test/actor.py').write('''
from __future__ import print_function
from leapp.actors import Actor
from leapp.models import TestModel
from leapp.tags import TestTag

class Test(Actor):
    name = 'test'
    description = 'No description has been provided for the test actor.'
    consumes = ()
    produces = (TestModel,)
    tags = (TestTag.Common,)

    def process(self):
        self.produce(TestModel(value='Some testing value'))
        if self._configuration:
            print(self.configuration.value)
''')
        check_call(['snactor', 'discover'])


def test_new_workflow(repository_dir):
    if repository_dir.join('workflows/test.py').check(file=True):
        # Test must have been run already
        return
    with repository_dir.as_cwd():
        check_call(['snactor', 'workflow', 'new', 'Test'])
        assert repository_dir.join('workflows/test.py').check(file=True)
        check_call(['snactor', 'discover'])
        content = repository_dir.join('workflows/test.py').read()
        repository_dir.join('workflows/test.py').write('from leapp.tags import TestTag\n' + content + '''

    class TestPhase(Phase):
         name = 'unit-test-workflow-phase'
         filter = TagFilter(TestTag)
         policies = Policies(Policies.Errors.FailPhase,
                             Policies.Retry.Phase)
         flags = Flags()
''')
        check_call(['snactor', 'discover'])


def test_run_workflow(repository_dir):
    # We need the workflow to be created already
    if not repository_dir.join('workflows/test.py').check(file=True):
        test_new_workflow(repository_dir)
    with pytest.raises(CalledProcessError):
        check_call(['snactor', 'workflow', 'run', 'Test'])
    with repository_dir.as_cwd():
        check_call(['snactor', 'workflow', 'run', 'Test'])
        check_call(['snactor', '--debug', 'workflow', 'run', 'Test'])
        check_call(['snactor', 'workflow', '--debug', 'run', 'Test'])
        check_call(['snactor', 'workflow', 'run', '--debug', 'Test'])
        test_clear_messages(repository_dir)
        connection = audit.create_connection('.leapp/leapp.db')
        assert not audit.get_messages(names=('TestModel',), context=context.last_snactor_context(connection),
                                      connection=connection)
        check_call(['snactor', 'workflow', 'run', '--debug', '--save-output', 'Test'])
        assert len(audit.get_messages(names=('TestModel',),
                                      context=context.last_snactor_context(connection), connection=connection)) == 1
        check_call(['snactor', 'workflow', 'run', 'Test'])
        assert len(audit.get_messages(names=('TestModel',),
                                      context=context.last_snactor_context(connection), connection=connection)) == 1
        check_call(['snactor', 'workflow', 'run', '--save-output', 'Test'])
        assert len(audit.get_messages(names=('TestModel',),
                                      context=context.last_snactor_context(connection), connection=connection)) == 2


def test_run_actor(repository_dir):
    with repository_dir.as_cwd():
        check_call(['snactor', 'run', 'Test'])
        check_call(['snactor', 'run', '--print-output', 'Test'])
        check_call(['snactor', 'run', '--save-output', 'Test'])
        check_call(['snactor', 'run', '--print-output', '--save-output', 'Test'])
        check_call(['snactor', 'run', '--debug', 'Test'])
        check_call(['snactor', 'run', '--debug', '--print-output', 'Test'])
        check_call(['snactor', 'run', '--debug', '--save-output', 'Test'])
        check_call(['snactor', 'run', '--debug', '--print-output', '--save-output', 'Test'])
        check_call(['snactor', '--debug', 'run', 'Test'])
        check_call(['snactor', '--debug', 'run', '--print-output', 'Test'])
        check_call(['snactor', '--debug', 'run', '--save-output', 'Test'])
        check_call(['snactor', '--debug', 'run', '--print-output', '--save-output', 'Test'])


def test_actor_config(repository_dir):
    with repository_dir.as_cwd():
        repository_dir.join('models/testconfig.py').write('''
from leapp.models import Model, fields
from leapp.topics import TestTopic


class TestConfig(Model):
    topic = TestTopic
    value = fields.String(default='Test config value')
''')
        check_call(['snactor', 'new-actor', 'ConfigTest'])
        repository_dir.join('actors/configtest/actor.py').write('''
from __future__ import print_function
from leapp.actors import Actor
from leapp.models import TestConfig
from leapp.tags import TestTag

class ConfigTest(Actor):
    name = 'test'
    description = 'No description has been provided for the test actor.'
    consumes = (TestConfig,)
    produces = ()
    tags = (TestTag,)

    def process(self):
        print(self.configuration.value)
''')
        with pytest.raises(CalledProcessError):
            # Raises if actor-config is not specified
            check_call(['snactor', 'run', 'ConfigTest'])
        # Inject message
        check_call(['snactor', 'messages', 'add', '-m', 'TestConfig', '{"value": "ConfigTest calls"}'])
        # Ensure the injected data is now available to the actor
        check_call(['snactor', 'run', '--actor-config', 'TestConfig', 'Test'])


def test_clear_messages(repository_dir):
    with repository_dir.as_cwd():
        check_call(['snactor', 'messages', 'clear'])



import datetime
import hashlib
import json
import multiprocessing
import os
import socket

from six.moves import configparser

from leapp.dialogs import RawMessageDialog
from leapp.dialogs.renderer import CommandlineRenderer
from leapp.exceptions import CannotConsumeErrorMessages
from leapp.messaging.answerstore import AnswerStore
from leapp.messaging.commands import WorkflowCommand
from leapp.models import DialogModel, ErrorModel
from leapp.utils import get_api_models


class BaseMessaging(object):
    """
    BaseMessaging is the Base class for all messaging implementations. It provides the basic interface that is
    supported within the framework. These are called the `produce` and `consume` methods.
    """

    def __init__(self, stored=True, config_model=None, answer_store=None):
        self._manager = multiprocessing.Manager()
        self._dialog_renderer = CommandlineRenderer()
        self._data = self._manager.list()
        self._answers = answer_store or AnswerStore(manager=self._manager)
        self._new_data = self._manager.list()
        self._commands = self._manager.list()
        self._errors = self._manager.list()
        self._stored = stored
        self._config_models = (config_model,) if config_model else ()
        self._dialogs = self._manager.list()
        self._stop_after_phase = self._manager.Value(bool, False)

    def load_answers(self, answer_file, workflow):
        """
        Loads answers from a given answer file

        :param answer_file: Path to file to load as answer file
        :type answer_file: str
        :param workflow: :py:class:`leapp.workflows.Workflow` instance to load the answers for.
        :type workflow: :py:class:`leapp.workflows.Workflow`
        :return: None
        """
        self._answers.load_and_translate_for_workflow(answer_file, workflow)

    @property
    def commands(self):
        """
        :return: List of commands that have been sent to the workflow execution.
        """
        return list(self._commands)

    @property
    def stored(self):
        """
        :return: If the messages are stored immediately, this function returns True, otherwise False.
        """
        return self._stored

    def dialogs(self):
        """
        Gets all dialogs actually encountered during workflow run

        :return: List of encountered dialogs
        """
        return list(self._dialogs)

    def errors(self):
        """
        Gets all produced errors.

        :return: List of newly produced errors
        """
        return list(self._errors)

    @property
    def stop_after_phase(self):
        """
        Returns True if execution stop was requested after the current

        :return: True if the executed was requested to be stopped.
        """
        return self._stop_after_phase.get()

    def messages(self):
        """
        Gets all newly produced messages.

        :return: List of newly processed messages
        """
        return list(self._new_data)

    def _perform_load(self, consumes):
        """
        Loads all messages that are requested from the `consumes` attribute of :py:class:`leapp.actors.Actor`

        :param consumes: Tuple or list of :py:class:`leapp.models.Model` types to preload
        :return: None
        """
        raise NotImplementedError()

    def _process_message(self, message):
        """
        This method performs the actual message sending, which can be sent over the network or stored
        in a database.

        :param message: The message data to process
        :type message: dict
        :return: Pass through a message that might get updated through the sending process.
        """
        raise NotImplementedError()

    def load(self, consumes):
        """
        Loads all messages that are requested from the `consumes` attribute of :py:class:`leapp.actors.Actor`

        :param consumes: Tuple or list of :py:class:`leapp.models.Model` types to preload
        :return: None
        :raises leapp.exceptions.CannotConsumeErrorMessages: When trying to consume ErrorModel
        """
        if ErrorModel in consumes:
            raise CannotConsumeErrorMessages()
        self._perform_load(consumes + self._config_models)

    def report_error(self, message, severity, actor, details):
        """
        Reports an execution error

        :param message: Message to print the error
        :type message: str
        :param severity: Severity of the error
        :type severity: leapp.models.error_severity.ErrorSeverity
        :param actor: Actor name that produced the message
        :type actor: leapp.actors.Actor
        :param details: A dictionary where additional context information can be passed along with the error
        :type details: dict
        :return: None
        """
        if details:
            details = json.dumps(details)
        model = ErrorModel(message=message, actor=actor.name, severity=severity, details=details,
                           time=datetime.datetime.now())
        self._do_produce(model, actor, self._errors)

    def request_stop_after_phase(self):
        """
        If called, it will cause the workflow to stop the execution after the current phase ends.
        """
        self._stop_after_phase.set(True)

    def register_dialog(self, dialog, actor):
        self._dialogs.append(dialog)
        userchoices = dialog.get_answers(self._answers)
        if not userchoices:
            # produce DialogModel messages for all the dialogs that don't have answers in answerfile
            self.produce(DialogModel(actor=actor.name, answerfile_sections=dialog.answerfile_sections), actor)
        else:
            # update dialogs with answers from answerfile. That is necessary for proper answerfile generation
            for component, value in userchoices.items():
                dialog_component = dialog.component_by_key(component)
                if dialog_component:
                    dialog_component.value = value

    def command(self, command):
        """
        Called to send a command to the workflow execution

        :param command: A command to send to the workflow execution.
        :type command: Instance of :py:class:`leapp.messaging.commands.WorkflowCommand`
        :return: None
        """
        if isinstance(command, WorkflowCommand):
            self._commands.append(command.encode())
        else:
            raise TypeError('Expected an instance of WorkflowCommand')

    def produce(self, model, actor):
        """
        Called to send a message available for other actors.

        :param model: Model to send as message payload
        :type model: :py:class:`leapp.models.Model`
        :param actor: Actor that sends the message
        :type actor: :py:class:`leapp.actors.Actor`
        :return: the updated message dict
        :rtype: dict
        """
        return self._do_produce(model, actor, self._new_data)

    def feed(self, model, actor):
        """
        Called to pre-fill sent messages and make them available for other actors.

        :param model: Model to send as message payload
        :type model: :py:class:`leapp.models.Model`
        :param actor: Actor that sends the message
        :type actor: :py:class:`leapp.actors.Actor`
        :return: the updated message dict
        :rtype: dict
        """
        return self._do_produce(model, actor, self._data, stored=False)

    def _do_produce(self, model, actor, target, stored=True):
        if not os.environ.get('LEAPP_HOSTNAME', None):
            os.environ['LEAPP_HOSTNAME'] = socket.getfqdn()
        data = json.dumps(model.dump(), sort_keys=True)
        message = {
            'type': type(model).__name__,
            'actor': type(actor).name,
            'topic': model.topic.name,
            'stamp': datetime.datetime.utcnow().isoformat() + 'Z',
            'phase': os.environ.get('LEAPP_CURRENT_PHASE', 'NON-WORKFLOW-EXECUTION'),
            'context': os.environ.get('LEAPP_EXECUTION_ID', 'TESTING-CONTEXT'),
            'hostname': os.environ['LEAPP_HOSTNAME'],
            'message': {
                'data': data,
                'hash': hashlib.sha256(data.encode('utf-8')).hexdigest()
            }
        }

        if stored and self.stored:
            self._process_message(message.copy())

        target.append(message)
        return message

    def get_answers(self, dialog):
        # NOTE(ivasilev): Non-Interactive. Returns answer from answerfile or empty dict straightaway.
        return dialog.get_answers(self._answers)

    def request_answers(self, dialog):
        # NOTE(ivasilev): Interactive. Should not be used in actors run at preupgrade\upgrade stage.
        #                 Will render the dialog and block further execution till user makes his choice.
        return dialog.request_answers(self._answers, self._dialog_renderer)

    def show_message(self, message):
        """
        Display a message in user interterface currently in use (CLI, GUI).

        It uses one of the dialog renderers in :py:mod:`leapp.dialogs.renderer`.

        :param message: Message to show
        :type message: str
        """
        RawMessageDialog(message=message).request_answers(self._answers, self._dialog_renderer)

    def consume(self, actor, *types):
        """
        Returns all consumable messages and filters them by `types`

        :param types: Variable number of :py:class:`leapp.models.Model` derived types to filter messages to be consumed
        :param actor: Actor that consumes the data
        :return: Iterable with messages matching the criteria
        """
        types = tuple((getattr(t, '_resolved', t) for t in types))
        messages = list(self._data) + list(self._new_data)
        # Needs to use get_api_models to consider all consumes including the one specified by Workflow APIs
        lookup = {model.__name__: model for model in get_api_models(type(actor), 'consumes') + self._config_models}
        if types:
            filtered = set(requested.__name__ for requested in types)
            messages = [message for message in messages if message['type'] in filtered]
        return (lookup[message['type']].create(json.loads(message['message']['data'])) for message in messages)



import errno
import json
import os
import re
import subprocess
import uuid

from leapp.exceptions import CommandError
from leapp.utils.clicmd import command_aware_wraps


def requires_repository(f):
    """
    Decorator for snactor commands that require to be run in a repository directory.
    """
    @command_aware_wraps(f)
    def checker(*args, **kwargs):
        if not find_repository_basedir('.'):
            raise CommandError('This command must be executed from the repository directory.')
        return f(*args, **kwargs)
    return checker


def to_snake_case(name):
    """
    Converts an UpperCaseName to a snake_case_name

    :param name: Name to convert
    :return: converted snake case name
    """
    if '-' in name:
        name = ''.join([part.capitalize() for part in name.split('-') if part])
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def make_class_name(name):
    """
    Converts a snake_case_name to an UpperCaseName

    :param name: Name to convert
    :return: Converted class name
    """
    return ''.join([x.capitalize() for x in to_snake_case(name).split('_')])


def make_name(name):
    """
    Converts a given name to a lower snake case

    :param name: Name to convert
    :return: Lower snake case
    """
    return to_snake_case(name)


def find_repository_basedir(path):
    """
    Tries to find the .leapp directory recursively ascending until it hits the root directory

    :param path: Path to start from (can be relative)
    :return: None if the base directory was not found, otherwise the absolute path to the base directory
    """
    path = os.path.realpath(path)
    while True:
        if os.path.isdir(os.path.join(path, '.leapp')):
            return path
        path, current = os.path.split(path)
        if not current:
            return None


def get_repository_metadata(path):
    """
    Gets the parsed metadata file as a dictionary

    :param path: Path to start loading the metadata from (it can be anywhere within the repository it will use
                 :py:func:`find_repository_dir` to find the repository base directory)
    :return: Dictionary with the metadata or an empty dictionary
    """
    basedir = find_repository_basedir(path)
    if basedir:
        with open(os.path.join(basedir, '.leapp', 'info'), 'r') as f:
            return json.load(f)
    return {}


def get_repository_name(path):
    """
    Retrieves the repository name from the repository metadata from within the given path. (it can be anywhere within
    the repository it will use :py:func:`find_repositoryt_dir` to find the repository base directory)
    :param path: Path within the leapp repository
    :return: Name of the repository
    :raises: KeyError if no name was found (e.g. not a valid repository path)
    """
    return get_repository_metadata(path)['name']


def _create_and_set_repository_id(path):
    data = get_repository_metadata(path)
    if data:
        basedir = find_repository_basedir(path)
        data['id'] = str(uuid.uuid4())
        with open(os.path.join(basedir, '.leapp', 'info'), 'w') as f:
            json.dump(data, f)
    return data.get('id')


def get_repository_links(path):
    """
    Retrieves a list of repository ids that are linked to given repository.

    :param path: Path within the leapp repository
    :return: List of repository ids this repository is linked to
    """
    return get_repository_metadata(path).get('repos', [])


def add_repository_link(path, repo_id):
    """
    Add a link from another repository to the current repository.

    :param path: Path within the leapp repository to modify
    :param repo_id: UUIDv4 string identifier for the repository to link
    :param repo_id: str
    :return: None
    """
    data = get_repository_metadata(path)
    if data and repo_id != data['id']:
        basedir = find_repository_basedir(path)
        data.setdefault('repos', []).append(repo_id)
        with open(os.path.join(basedir, '.leapp', 'info'), 'w') as f:
            json.dump(data, f)
            return True
    return False


def get_repository_id(path):
    """
    Retrieves the repository name from the repository metadata from within the given path. (it can be anywhere within
    the repository it will use :py:func:`find_repository_dir` to find the repository base directory)
    :param path: Path within the leapp repository
    :return: ID of the repository
    :raises: KeyError if no name was found (e.g. not a valid repository path)
    """
    try:
        return get_repository_metadata(path)['id']
    except KeyError:
        if get_repository_metadata(path):
            return _create_and_set_repository_id(path)
        raise


def get_user_config_path():
    """
    Returns the path to the user configuration directory and creates it if it does not exist already.

    :return: Path to the configuration directory of leapp.
    """
    leapp_conf = os.path.join(os.path.expanduser('~'), '.config', 'leapp')
    try:
        os.makedirs(leapp_conf)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise
    return leapp_conf


def get_user_config_repos():
    """
    Returns the path to the user config file for repositories.

    :return: Path to the repos.json configuration file.
    """
    return os.path.join(get_user_config_path(), 'repos.json')


def get_user_config_repo_data():
    """
    Returns the user config repository data.

    :return: Data for user configurations.
    """
    repos = get_user_config_repos()
    if not os.path.isfile(repos):
        return {}
    with open(repos) as f:
        return json.load(f)


def find_repos(path):
    """
    Finds repositories within the given path.

    :param path: Path to search for repositories.
    :return: List of strings with found repository paths.
    """
    try:
        with open(os.devnull, 'w') as devnull:
            data = subprocess.check_output(
                ['/usr/bin/find', '-L', path, '-name', '.leapp'], stderr=devnull).decode('utf-8').split('\n')
            return [os.path.abspath(os.path.dirname(rpath)) for rpath in data if rpath.strip()]
    except subprocess.CalledProcessError:
        return ()


def get_global_repositories_data():
    """
    Returns the data of all system wide available repositories.

    :return: Repository information
    """
    enabled = {os.path.realpath(path) for path in find_repos('/etc/leapp/repos.d') if path.strip()}
    all_repos = {os.path.realpath(path) for path in find_repos('/usr/share/leapp-repository') if path.strip()}
    repo_data = {}
    for repo in all_repos:
        repo_id = get_repository_metadata(repo).get('id', None)
        if not repo_id:
            continue
        repo_data[repo_id] = {
            'id': repo_id,
            'path': repo,
            'name': get_repository_name(repo),
            'enabled': repo in enabled
        }
    return repo_data



import os
import sys

from leapp.utils.repository import requires_repository, make_class_name, make_name, find_repository_basedir
from leapp.utils.clicmd import command, command_arg, command_opt
from leapp.exceptions import CommandError


_LONG_DESCRIPTION = '''
Creates a new Actor with all necessary boilerplate code in the current
repository.

The options --tags, --consumes, and --produces can be used multiple times, for multiple values.
Example:
  $ snactor new-actor ExampleActor --tag SomeWorkflowTag --tag AnotherTag --produces ImportantModel


For more information please consider reading the documentation at:
https://red.ht/leapp-docs
'''


def quoted(x):
    return "'" + x + "'"


def as_quoted_tuple(values):
    if values:
        suffix = ''
        if len(values) == 1:
            suffix = ','
        return ', '.join(values) + suffix
    return ''


@command('new-actor', help='Creates a new actor', description=_LONG_DESCRIPTION)
@command_arg('actor-name')
@command_opt('--tag', action='append', metavar='TagClassName', help='Existing Tag to add to the tags field')
@command_opt('--consumes', action='append', metavar='ModelClassName',
             help='Existing Model to add to the consumes field')
@command_opt('--produces', action='append', metavar='ModelClassName',
             help='Existing Model to add to the produces field')
@requires_repository
def cli(args):
    actor_name = args.actor_name
    basedir = find_repository_basedir('.')

    tag_imports = ''
    model_imports = ''
    if args.tag:
        tag_imports = '\nfrom leapp.tags import {}'.format(', '.join(tuple([x.split('.')[0] for x in args.tag])))
    if args.consumes or args.produces:
        models = set((args.produces or []) + (args.consumes or []))
        model_imports = '\nfrom leapp.models import {}'.format(', '.join(models))

    tags_content = '({})'.format(as_quoted_tuple(args.tag))
    consumes_content = '({})'.format(as_quoted_tuple(args.consumes))
    produces_content = '({})'.format(as_quoted_tuple(args.produces))

    actors_dir = os.path.join(basedir, 'actors')
    if not os.path.isdir(actors_dir):
        os.mkdir(actors_dir)

    actor_dir = os.path.join(actors_dir, actor_name.lower())
    if not os.path.isdir(actor_dir):
        os.mkdir(actor_dir)

    actor_test_dir = os.path.join(actor_dir, 'tests')
    if not os.path.isdir(actor_test_dir):
        os.mkdir(actor_test_dir)

    actor_path = os.path.join(actor_dir, 'actor.py')
    if os.path.exists(actor_path):
        raise CommandError("File already exists: {}".format(actor_path))

    with open(actor_path, 'w') as f:
        f.write('''from leapp.actors import Actor{model_imports}{tag_imports}


class {actor_class}(Actor):
    """
    No documentation has been provided for the {actor_name} actor.
    """

    name = '{actor_name}'
    consumes = {consumes_content}
    produces = {produces_content}
    tags = {tags_content}

    def process(self):
        pass
'''.format(actor_class=make_class_name(actor_name), actor_name=make_name(actor_name), tags_content=tags_content,
           produces_content=produces_content, consumes_content=consumes_content,
           model_imports=model_imports, tag_imports=tag_imports))

    sys.stdout.write("New actor {} has been created at {}\n".format(make_class_name(actor_name),
                                                                    os.path.realpath(actor_path)))
    return os.path.dirname(os.path.realpath(actor_path))



from __future__ import print_function

import datetime
import os
import uuid

import py
import pytest

import leapp.utils.audit
from leapp.messaging.inprocess import InProcessMessaging
from leapp.repository.scan import scan_repo


@pytest.fixture(scope='module')
def repository(leapp_forked):  # noqa; pylint: disable=unused-argument
    repository_path = py.path.local(
        os.path.join(os.path.dirname(os.path.dirname(__file__)), 'data', 'deprecation-tests'))
    with repository_path.as_cwd():
        repo = scan_repo('.')
        repo.load(resolve=True)
        yield repo


def test_deprecations(repository):
    start = datetime.datetime.utcnow()
    actor = repository.lookup_actor('DeprecationTests')
    messaging = InProcessMessaging()
    os.environ['LEAPP_EXECUTION_ID'] = str(uuid.uuid4())
    with py.path.local(actor.directory).as_cwd():
        actor(messaging=messaging).run()
    entries = leapp.utils.audit.get_audit_entry('deprecation', os.getenv('LEAPP_EXECUTION_ID'))
    assert entries
    entries = [entry for entry in entries
               if datetime.datetime.strptime(entry['stamp'].rstrip('Z'), '%Y-%m-%dT%H:%M:%S.%f') > start]
    assert entries and len(entries) == 5



from leapp.utils.meta import with_metaclass
from leapp.workflows.flags import Flags
from leapp.workflows.policies import Policies


class PhaseMeta(type):
    classes = []

    def __new__(mcs, name, bases, attrs):
        klass = super(PhaseMeta, mcs).__new__(mcs, name, bases, attrs)
        PhaseMeta.classes.append(klass)
        return klass


class Phase(with_metaclass(PhaseMeta)):
    name = None
    filter = None
    policies = Policies(Policies.Errors.FailPhase,
                        Policies.Retry.Phase)
    flags = Flags()

    @classmethod
    def get_index(cls):
        return PhaseMeta.classes.index(cls)

    @classmethod
    def serialize(cls):
        """
        :return: Dictionary with the serialized representation of the phase
        """
        return {
            'name': cls.name,
            'class_name': cls.__name__,
            'index': cls.get_index(),
            'filter': cls.filter.serialize() if cls.filter else None,
            'policies': cls.policies.serialize(),
            'flags': cls.flags.serialize(),
        }



from leapp.exceptions import CyclingDependenciesError


class PhaseActors(object):
    def __init__(self, actors, stage):
        self.stage = stage
        self._actors = actors
        self._consumes = set()
        self._produces = set()
        self._messages = {}

        for actor in self._actors:
            self._consumes.update(actor.consumes)
            self._produces.update(actor.produces)
            for message in actor.produces:
                self._messages.setdefault(message.__name__,
                                          {'type': message, 'producers': []})['producers'].append(actor)
            for message in actor.consumes:
                self._messages.setdefault(message.__name__, {'type': message, 'producers': []})
        self._initial = self._consumes - self._produces
        self._sort()

    @property
    def initial(self):
        return tuple(self._initial)

    @property
    def actors(self):
        return tuple(self._actors)

    @property
    def consumes(self):
        return tuple(self._consumes)

    @property
    def produces(self):
        return tuple(self._produces)

    def _sort(self):
        actors, self._actors = list(self._actors), ()
        while actors:
            scheduled = []
            for idx, actor in enumerate(actors):
                for message in actor.consumes:
                    if self._messages[message.__name__]['producers']:
                        break
                else:
                    for message in actor.produces:
                        self._messages[message.__name__]['producers'].remove(actor)

                    scheduled.append(idx)
                    self._actors += (actor,)

            if not scheduled:
                raise CyclingDependenciesError(
                    "Could not solve dependency order for '{}'".format(', '.join([actor.name for actor in actors])))

            for idx in reversed(scheduled):
                actors.pop(idx)



from leapp.utils.deprecation import deprecated


@deprecated('2010-01-01', 'Deprecated class without __init__')
class DeprecatedNoInit(object):
    pass


@deprecated('2010-01-01', 'Deprecated class with __init__')
class DeprecatedWithInit(object):
    pass


@deprecated('2010-01-01', 'Deprecated base class without __init__')
class DeprecatedBaseNoInit(object):
    pass


class DeprecatedNoInitDerived(DeprecatedBaseNoInit):
    pass


@deprecated('2010-01-01', 'Deprecated base class with __init__')
class DeprecatedBaseWithInit(object):
    def __init__(self):
        pass


class DeprecatedWithInitDerived(DeprecatedBaseWithInit):
    def __init__(self):
        pass


@deprecated('2010-01-01', 'This function is no longer supported.')
def deprecated_function():
    pass



"""
:py:mod:`leapp.libraries.stdlib`
represents a location for functions that otherwise would be defined multiple times across leapp actors
and at the same time, they are really useful for other actors.
"""
import logging
import os
import sys
import uuid
import base64

from leapp.exceptions import LeappError
from leapp.utils.audit import create_audit_entry
from leapp.libraries.stdlib import api
from leapp.libraries.stdlib.call import _call, STDOUT
from leapp.libraries.stdlib.config import is_debug


class CalledProcessError(LeappError):
    """
    Leapp Call Process Exception Error.

    Raised when the result of a called process is of a none zero return code.
    """
    def __init__(self, message, command, result):
        """
        Initialize CalledProcessError Exception.
        :param message: An CalledProcessError exception message.
        :param command: The command that has been executed, with its arguments.
        :param result: A non-zero whatever result that the command returned.
        """
        super(CalledProcessError, self).__init__(message)
        self._result = result
        self.command = command

    @property
    def stdout(self):
        """
        Retrieve the stdout.
        :return: Standard Output.
        """
        return self._result.get('stdout')

    @property
    def stderr(self):
        """
        Retrieve the stderr.
        :return: Standard Error.
        """
        return self._result.get('stderr')

    @property
    def exit_code(self):
        """
        Retrieve the exit code.
        :return: An exit code.
        """
        return self._result.get('exit_code')

    @property
    def signal(self):
        """
        Retrieve the signal which the process was signalled by.
        :return: A signal that the process received.
        """
        return self._result.get('signal')

    @property
    def pid(self):
        """
        Retrieve the pid of the finished process.
        :return: The pid of the process.
        """
        return self._result.get('pid')


# FIXME: Issue #488
def __write_raw(fd_info, a_buffer):
    """
    Raw write to fd compatible in Py2 and Py3 (3.3+)

    Buffer Output Discrimination is based on fd_type in fd_info.

    :param fd_info: Contains File descriptor type
    :type fd_info: tuple
    :param a_buffer: buffer interface
    :type a_buffer: bytes array
    """
    (unused_fd, fd_type) = fd_info
    if sys.version_info > (3, 0):
        if fd_type == STDOUT:
            fd = sys.stdout.fileno()
        else:
            # FIXME: Issue #488 - what to do in case fd_type != STDERR?
            fd = sys.stderr.fileno()
        os.writev(fd, [a_buffer])
    else:
        if fd_type == STDOUT:
            sys.stdout.write(a_buffer)
        else:
            sys.stderr.write(a_buffer)


# FIXME: Issue #488
def _console_logging_handler(fd_info, a_buffer):
    """
    Log into either STDOUT or to STDERR.

    Buffer Output Discrimination is based on fd_type in fd_info while in a leapp debug mode.

    :param fd_info: Contains File descriptor type
    :type fd_info: tuple
    :param a_buffer: buffer interface
    :type a_buffer: bytes array
    """
    if is_debug():
        __write_raw(fd_info, a_buffer)


def _logfile_logging_handler(fd_info, line):  # pylint: disable=unused-argument
    """
    Log into logfile

    :param fd_info: Contains File descriptor type (not used)
    :type fd_info: tuple
    :param line: data to be printed to a logfile
    :type line: string
    """
    def _raise_level(handlers):
        for handler in handlers:
            handler.setLevel(handler.level + 1)

    def _restore_level(handlers, levels):
        for handler, level in zip(handlers, levels):
            handler.setLevel(level)

    if is_debug():
        # raise logging level of stream handlers above DEBUG to avoid duplicated output to stdout/stderr
        logger_handlers = api.current_logger().root.handlers
        stream_handlers = [handler for handler in logger_handlers if isinstance(handler, logging.StreamHandler)]
        handler_levels = [handler.level for handler in stream_handlers]
        _raise_level(stream_handlers)
        api.current_logger().debug(line)
        _restore_level(stream_handlers, handler_levels)
    else:
        api.current_logger().debug(line)


def run(args, split=False, callback_raw=_console_logging_handler, callback_linebuffered=_logfile_logging_handler,
        env=None, checked=True, stdin=None, encoding='utf-8'):
    """
    Run a command and return its result as a dict.

    The execution of the program and it's results are captured by the audit.

    :param args: Command to execute
    :type args: list or tuple
    :param split: Split the output on newlines
    :type split: bool
    :param callback_raw: Optional custom callback executed on raw data to print in console
    :type callback_raw: (fd: int, buffer: bytes) -> None
    :param env: Environment variables to use for execution of the command
    :type env: dict
    :param checked: Raise an exception on a non-zero exit code, default True
    :type checked: bool
    :param stdin: String or a file descriptor that will be written to stdin of the child process
    :type stdin: int, str
    :return: {'stdout' : stdout, 'stderr': stderr, 'signal': signal, 'exit_code': exit_code, 'pid': pid}
    :rtype: dict
    """
    if not args:
        message = 'Command to call is missing.'
        api.current_logger().error(message)
        raise ValueError(message)
    api.current_logger().debug('External command has started: {0}'.format(str(args)))
    _id = str(uuid.uuid4())
    result = None
    try:
        create_audit_entry('process-start', {'id': _id, 'parameters': args, 'env': env})
        result = _call(args, callback_raw=callback_raw, callback_linebuffered=callback_linebuffered,
                       stdin=stdin, env=env, encoding=encoding)
        if checked and result['exit_code'] != 0:
            message = 'Command {0} failed with exit code {1}.'.format(str(args), result.get('exit_code'))
            api.current_logger().debug(message)
            raise CalledProcessError(
                message=message,
                command=args,
                result=result
            )
        if split and encoding:
            result.update({
                'stdout': result['stdout'].splitlines()
            })
    finally:
        audit_result = result
        if not encoding:
            audit_result = result.copy()
            audit_result.update({
                'stdout': 'Base64: ' + base64.b64encode(result['stdout']).decode('utf-8'),
                'stderr': 'Base64: ' + base64.b64encode(result['stderr']).decode('utf-8')
            })
        create_audit_entry(
            'process-result', {'id': _id, 'parameters': args, 'result': audit_result, 'env': env}
        )
        api.current_logger().debug('External command has finished: {0}'.format(str(args)))
    return result



import os

import sys

from leapp.utils.repository import (make_class_name, make_name, find_repository_basedir,
                                    requires_repository)
from leapp.utils.clicmd import command_arg, command
from leapp.exceptions import CommandError

_LONG_DESCRIPTION = '''
Creates a new Topic in the current repository.

For more information please consider reading the documentation at:
https://red.ht/leapp-docs
'''


@command('new-topic', help='Creates a new topic')
@command_arg('topic-name')
@requires_repository
def cli(args):
    topic_name = args.topic_name
    basedir = find_repository_basedir('.')

    basedir = os.path.join(basedir, 'topics')
    if not os.path.isdir(basedir):
        os.mkdir(basedir)

    topic_path = os.path.join(basedir, topic_name.lower() + '.py')
    if os.path.exists(topic_path):
        raise CommandError("File already exists: {}".format(topic_path))

    topic_path = os.path.join(basedir, topic_name.lower() + '.py')
    topic_class_name = make_class_name(topic_name)
    if not topic_class_name.endswith('Topic'):
        topic_class_name += 'Topic'
    with open(topic_path, 'w') as f:
        f.write('''from leapp.topics import Topic


class {topic_name}(Topic):
    name = '{topic}'
'''.format(topic_name=topic_class_name, topic=make_name(topic_name)))

    sys.stdout.write("New topic {} has been created in {}\n".format(topic_class_name, os.path.realpath(topic_path)))



import sys
from subprocess import check_call
from setuptools import find_packages, setup
from distutils.util import convert_path

main_ns = {}
ver_path = convert_path('leapp/__init__.py')
with open(ver_path) as ver_file:
    exec(ver_file.read(), main_ns)

check_call([sys.executable, 'res/schema/embed.py'])

with open('res/schema/schemas.py', 'r') as orig:
    with open('leapp/utils/schemas.py', 'w') as target:
        target.write(orig.read())

EXCLUSION = []
if sys.version_info[0] > 2:
    # Python 2 only
    EXCLUSION.append('leapp.compatpy2only')

setup(
    name='leapp',
    version=main_ns['VERSION'],
    packages=find_packages(exclude=EXCLUSION),
    install_requires=['six', 'requests'],
    entry_points='''
        [console_scripts]
        snactor=leapp.snactor:main
        leapp=leapp.cli:main

        [pytest11]
        snactor_plugin=leapp.snactor.fixture
    '''
)



import functools
import os
import re

from six.moves import configparser

from leapp.utils.repository import find_repository_basedir


_LEAPP_CONFIG = None

# files that will get reported at the end of a preupgrade/upgrade run if they were created/modified during it
_REPORTS = [
    'leapp-report.json',
    'leapp-report.txt',
]

# debug logs that will get reported at the end of a preupgrade/upgrade run if they were created/modified during it
_LOGS = [
    'leapp-upgrade.log',
    'leapp-preupgrade.log'
]

# files that will go into the leapp-logs.tar.gz archive and get deleted each time a run is started
_FILES_TO_ARCHIVE = [
    'dnf-plugin-data.txt',
] + _REPORTS + _LOGS

_CONFIG_DEFAULTS = {
    'archive': {
        'dir': '/var/log/leapp/archive/',
    },
    'database': {
        'path': '/var/lib/leapp/leapp.db',
    },
    'debug': {
        'dir': '/var/log/leapp/dnf-debugdata/',
    },
    'files_to_archive': {
        'dir': '/var/log/leapp/',
        'files': ','.join(_FILES_TO_ARCHIVE),
    },
    'logs': {
        'dir': '/var/log/leapp/',
        'files': ','.join(_LOGS),
    },
    'report': {
        'dir': '/var/log/leapp/',
        'files': ','.join(_REPORTS),
        'answerfile': '/var/log/leapp/answerfile',
        'userchoices': '/var/log/leapp/answerfile.userchoices'
    },
    'repositories': {
        'repo_path': '.',
    },
}


class BetterConfigParser(configparser.ConfigParser):
    _EINTERPOL = re.compile(r"\$\{([^}]+)\}")

    def get(self, section, *args, **kwargs):
        depth = kwargs.pop('depth', configparser.MAX_INTERPOLATION_DEPTH)
        return self._resolve(BetterConfigParser.__bases__[0].get(self, section, *args, **kwargs),
                             depth=depth,
                             section=section)

    def _resolve(self, value, depth, section):
        if depth:
            if '${' not in value:
                return value
            replacer = functools.partial(self._sub_interpol, depth=depth, section=section)
            value = BetterConfigParser._EINTERPOL.sub(replacer, value)
        return value

    def _sub_interpol(self, match, depth, section):
        s = match.group(1)
        if not s:
            return match.group()
        parts = s.split(':')
        if len(parts) == 1:
            parts.insert(0, section)
        return self.get(*parts, depth=depth - 1)


def get_config():
    global _LEAPP_CONFIG
    if not _LEAPP_CONFIG:
        repository_defaults = {}
        if find_repository_basedir(os.environ.get('LEAPP_CONFIG', '.')):
            repository_defaults['repository'] = {
                'root_dir': find_repository_basedir(os.environ.get('LEAPP_CONFIG', '.')),
                'state_dir': '${root_dir}/.leapp',
            }
            # Backwards compatibility for older repositories that still used the 'project' terminology.
            repository_defaults['project'] = repository_defaults['repository']
        _LEAPP_CONFIG = BetterConfigParser()
        for section, values in tuple(_CONFIG_DEFAULTS.items()) + tuple(repository_defaults.items()):
            if not _LEAPP_CONFIG.has_section(section):
                _LEAPP_CONFIG.add_section(section)
            for name, value in values.items():
                if value is not None:
                    _LEAPP_CONFIG.set(section, name, value)
        _LEAPP_CONFIG.read([os.getenv('LEAPP_CONFIG', '/etc/leapp/leapp.conf')])

    return _LEAPP_CONFIG



from leapp.actors import Actor
from leapp.exceptions import StopActorExecutionError
from leapp.models import DepCheck2, DepCheck4
from leapp.tags import FirstPhaseTag, WorkflowApiTestWorkflowTag


class DepCheckConsumes(Actor):
    """
    This actor consumes the Models DepCheck2 and DepCheck4 produced by the DepCheck actor to verify the functionality.
    """

    name = 'dep_check_consumes'
    produces = ()
    consumes = (DepCheck2, DepCheck4)
    tags = (FirstPhaseTag, WorkflowApiTestWorkflowTag)

    def process(self):
        # Only one message is supposed to be produced
        if len(list(self.consume(DepCheck2))) != 1:
            raise StopActorExecutionError(
                "Expected 1 DepCheck2 message => {}".format(len(list(self.consume(DepCheck2)))))

        # Only one message is supposed to be produced
        if len(list(self.consume(DepCheck4))) != 1:
            raise StopActorExecutionError(
                "Expected 1 DepCheck4 message => {}".format(len(list(self.consume(DepCheck4)))))



from leapp.dialogs.components import TextComponent, PasswordComponent, NumberComponent, BooleanComponent, \
    ChoiceComponent, MultipleChoiceComponent
from leapp.dialogs.dialog import Dialog
from leapp.dialogs.renderer import CommandlineRenderer


class UsernamePasswordDialog(Dialog):
    """ Reusable username and password request dialog. """
    title = 'Please enter username and password'
    components = (
        TextComponent(key='username', label='Username', default=None),
        PasswordComponent(key='password', label='Password', default=None)
    )


class RawMessageDialog(Dialog):
    """ Reusable message dialog. - This sends a message to the user only without any special formatting. """
    title = ''

    def __init__(self, message):
        super(RawMessageDialog, self).__init__(scope=None, reason=message)

    def request_answers(self, store, renderer):
        """
        :param store: AnswerStore instance
        :param renderer: Target renderer instance
        :return: Dictionary with answers once retrieved
        """
        renderer.render(self)
        return {}



import subprocess


def reboot_system():
    subprocess.Popen(['/sbin/shutdown', '-r', 'now'])


def get_api_models(actor, what):
    """
    Used to retrieve the full list of models including the ones defined by WorkflowAPIs used by the actor.

    :param what: A string which either is 'consumes' or 'produces'
    :type what: str
    :param actor: Actor type/instance or ActorDefinition instance to retrieve the information from
    :type actor: Actor or ActorDefinition
    :return: Tuple of all produced or consumed models as specified by actor and APIs used by the actor.
    """

    def _enforce_tuple(x):
        if not isinstance(x, tuple):
            return (x,)
        return x

    def _do_get(api):
        result = _enforce_tuple(getattr(api, what, ()))
        for a in _enforce_tuple(api.apis or ()):
            result = result + _do_get(a)
        return result
    return tuple(set(_do_get(actor)))



import pytest

from leapp.exceptions import InvalidTopicDefinitionError
from leapp.topics import Topic, ErrorTopic, get_topics
import leapp.topics


class UnitTestTopic(Topic):
    name = "test-topic"


def test_topic():
    assert getattr(leapp.topics, 'UnitTestTopic') is UnitTestTopic


def test_get_topics():
    topics = get_topics()
    assert len(topics) >= 2
    assert UnitTestTopic in topics
    assert ErrorTopic in topics
    UnitTestTopic.name = None
    with pytest.raises(InvalidTopicDefinitionError):
        get_topics()



from leapp.workflows.api import WorkflowAPI
from leapp.workflows.api import v2


class TestAPI(WorkflowAPI):
    """
    V1 of the TestAPI it also depends on the v2.TestAPI
    """
    apis = (v2.TestAPI,)

    def __init__(self):
        # v2 API instance
        self._impl = v2.TestAPI()

    def order_change(self, a, b, c):
        """
        Returns a tuple in the form of a, b, c
        """
        print("v1.TestAPI.order_change => " + str((a, b, c)))
        # Calls v2 api implementation with the corrected parameter order
        # which then calls the v3 api with the corrected parameter order
        # returning the value received
        return self._impl.order_change(c, b, a)

    def survivor(self, value):
        """
        Prefixes a given value - Passed through to the latest version
        """
        print("v1.TestAPI.survivor")
        # Calls v2 api implementation which then calls the v3 api
        # returning the value received
        return self._impl.survivor(message=value)

    def removed(self, yadda):
        print("v1.TestAPI.removed('{}')".format(yadda))
        print("I am removed in newer versions")
        # Does not really do anything as it is removed in v2 and higher



import datetime
import json
import os
import sys
from importlib import import_module

from leapp.exceptions import CommandError, LeappError
from leapp.logger import configure_logger
from leapp.messaging.inprocess import InProcessMessaging
from leapp.repository.scan import find_and_scan_repositories
from leapp.snactor.context import with_snactor_context
from leapp.utils.clicmd import command, command_arg, command_opt
from leapp.utils.output import beautify_actor_exception, report_deprecations, report_errors
from leapp.utils.repository import find_repository_basedir, requires_repository

_LONG_DESCRIPTION = '''
Runs the given actor as specified as `actor_name` in a testing environment.

For more information please consider reading the documentation at:
https://red.ht/leapp-docs
'''


@command('run', help='Execute the given actor', description=_LONG_DESCRIPTION)
@command_arg('actor-name')
@command_opt('--actor-config', help='Name of the workflow config model to use')
@command_opt('--save-output', is_flag=True, help='Save the produced messages by this actor.')
@command_opt('--print-output', is_flag=True, help='Print the produced messages by this actor.')
@requires_repository
@with_snactor_context
def cli(args):
    start = datetime.datetime.utcnow()
    log = configure_logger()
    basedir = find_repository_basedir('.')
    repository = find_and_scan_repositories(basedir, include_locals=True)
    try:
        repository.load()
    except LeappError as exc:
        sys.stderr.write(exc.message)
        sys.stderr.write('\n')
        sys.exit(1)
    actor_logger = log.getChild('actors')
    actor = repository.lookup_actor(args.actor_name)
    if not actor:
        raise CommandError('Actor "{}" not found!'.format(args.actor_name))
    config_model = getattr(import_module('leapp.models'), args.actor_config) if args.actor_config else None
    messaging = InProcessMessaging(stored=args.save_output, config_model=config_model)
    messaging.load(actor.consumes)

    failure = False
    with beautify_actor_exception():
        try:
            actor(messaging=messaging, logger=actor_logger, config_model=config_model).run()
        except BaseException:
            failure = True
            raise

    report_errors(messaging.errors())
    report_deprecations(os.getenv('LEAPP_EXECUTION_ID'), start=start)

    if failure or messaging.errors():
        sys.exit(1)

    if args.print_output:
        json.dump(messaging.messages(), sys.stdout, indent=2)
        sys.stdout.write('\n')
